from .app import App
from .geometry import Geometry
from .rois.roi_data import RoiData
from .rois.roi import Roi, RoiTypes
from .fitting import FitObject
from .debug_tracker import TrackQObjects
from .update import launch_detached



from typing import Dict, Set
from collections import defaultdict

from h5py import Group

from ..file_manager import ImageKey, FolderKey
from .roi_data import RoiData


class RoiMetaData(object):

    H5_NAME = 'roi_metadata'

    def __init__(self, folder_key: FolderKey, *,
                 names: Dict[int, str] = None, groups: Dict[int, str] = None):

        self.folder_key = folder_key.clean_copy()
        self.names: Dict[int, str] = names or {}
        self.groups: Dict[int, str] = groups or {}
        self.keys: Dict[int, Set[ImageKey]] = defaultdict(set)

    def update_metadata(self, roi_data: RoiData, image_key: ImageKey):
        for key in roi_data.keys():
            self.keys[key].add(image_key)

    def to_h5(self, h5_group: Group):
        pass

    @classmethod
    def from_h5(cls, folder_key: FolderKey, h5_group: Group):
        # TODO implement read/save for h5 file
        return cls(folder_key)

    @staticmethod
    def del_h5(h5_group: Group):
        if RoiMetaData.H5_NAME in h5_group.keys():
            del h5_group[RoiMetaData.H5_NAME]



from enum import Enum

import numpy as np

__all__ = ['TransformationsHolder', 'Transformation', 'UnknownTransformation']


class UnknownTransformation(ValueError):
    pass


class Flip(object):
    __slots__ = ('axis',)

    def __init__(self, axis: int):
        self.axis = axis

    def __call__(self, img: np.ndarray):
        return np.flip(img, self.axis)


class Rotate(object):
    __slots__ = ('k',)

    def __init__(self, k: int):
        self.k = k

    def __call__(self, img: np.ndarray):
        return np.rot90(img, self.k)


class Transformation(Enum):
    horizontal_flip = Flip(1)
    vertical_flip = Flip(0)
    rotate_right = Rotate(-1)
    rotate_left = Rotate(1)


_T_DICT = {
    '1234': (),
    '1324': (Flip(1), Rotate(1)),
    '2413': (Rotate(1),),
    '2143': (Flip(1),),
    '3142': (Rotate(-1),),
    '3412': (Flip(0),),
    '4321': (Flip(0), Flip(1)),
    '4231': (Rotate(1), Flip(1))
}


class TransformationsHolder(object):
    def __init__(self, key: str = None):
        self._img = np.array([[1, 2], [3, 4]])
        if key:
            self.update(key)

    def __call__(self, image):
        for op in self._operations:
            image = op(image)
        return image

    def add(self, op: Transformation):
        self._img = op.value(self._img)

    def clear(self):
        self._img = np.array([[1, 2], [3, 4]])

    @property
    def _operations(self) -> tuple:
        return _T_DICT[self.key]

    @property
    def key(self):
        return ''.join(map(str, self._img.ravel()))

    def update(self, key: str):
        self._img = np.array([[1, 2], [3, 4]])
        try:
            for op in _T_DICT[key]:
                self._img = op(self._img)
        except KeyError:
            raise UnknownTransformation(f'Key {key} doesn\'t correspond to any known transformation.')



import logging
from pathlib import Path
from typing import List, Union
import re
import weakref
from abc import abstractmethod
from copy import deepcopy

from ..read_image import read_image

from h5py import File, Group, Dataset

AVAILABLE_IMAGE_FORMATS = tuple('.tif .tiff .edf .edf.gz'.split())
GLOB_IMAGE_FORMATS = 'edf, tiff, h5 files (*.tiff *.edf *.tif *.edf.gz *.h5 *.hdf5)'
H5_FORMAT = tuple('.h5 .hdf5'.split())

PROJECT_KEY = '__GIWAXS_PROJECT__'
IMAGE_PROJECT_KEY = '__GIWAXS_IMAGE_GROUP__'

logger = logging.getLogger(__name__)


def _check_project(h5path: Path) -> bool or None:
    try:
        with File(str(h5path.resolve()), 'r') as f:
            if PROJECT_KEY in f.attrs.keys():
                return True
            else:
                return False
    except Exception as err:
        logger.exception(err)
        return


def _file_name(key: str, name: str = ''):
    return f'{key}{"_" if name else ""}{name}.giwaxs'


class InvalidKey(Exception):
    """The corresponding data does not exist or corrupted."""


class AbstractKey(object):
    is_project = False

    def __init__(self, parent=None, **kwargs):
        self._parent = weakref.ref(parent) if parent else None

    def file_name(self, name: str = '') -> str:
        return re.sub('[^\w\-_\. ]', '_', _file_name(self._file_key(), name))

    @abstractmethod
    def _file_key(self) -> str:
        pass

    @property
    def name(self):
        return

    @property
    def parent(self) -> 'AbstractKey' or None:
        return self._parent() if self._parent else None

    def remove_parent(self) -> 'AbstractKey' or None:
        parent = self.parent
        self._parent = None
        return parent

    def set_parent(self, parent: 'AbstractKey'):
        self._parent = weakref.ref(parent)

    @abstractmethod
    def __eq__(self, other):
        pass

    @abstractmethod
    def is_valid(self) -> bool:
        pass


class FolderKey(AbstractKey):
    def __init__(self, parent: 'FolderKey', **kwargs):
        super().__init__(parent, **kwargs)
        self._image_children: List[ImageKey] = []
        self._folder_children: List[FolderKey] = []
        self._updated: bool = False

    def is_updated(self) -> bool:
        return self._updated

    @property
    def image_children(self):
        yield from self._image_children

    @property
    def folder_children(self):
        yield from self._folder_children

    @property
    def images_num(self):
        return len(self._image_children)

    @property
    def folders_num(self):
        return len(self._folder_children)

    def update(self):
        self.clear()
        self._updated = True

    def clear(self):
        self._image_children: List[ImageKey] = []
        self._folder_children: List[FolderKey] = []

    def detach_subfolders(self) -> List['FolderKey']:
        subfolders, self._folder_children = self._folder_children, []
        return subfolders

    def attach_subfolders(self, subfolders):
        self._folder_children = subfolders

    def clean_copy(self):
        with RemoveWeakrefs(self, restore=True, remove_subfolders=True):
            return deepcopy(self)

    def image_idx(self, key: 'ImageKey'):
        try:
            return self._image_children.index(key)
        except ValueError:
            return

    def image_by_key(self, idx: int):
        try:
            return self._image_children[idx]
        except IndexError:
            return

    def get_next_image(self, key: 'ImageKey') -> 'ImageKey' or None:
        if not self._image_children:
            return
        idx = key.idx
        if idx is None:
            idx = key.idx = self.image_idx(key)
            if idx is None:
                return
        if idx + 1 < len(self._image_children):
            return self._image_children[idx + 1]

    def remove_image(self, key: 'ImageKey'):
        idx = self.image_idx(key)
        if idx is None:
            return
        self._image_children.remove(key)
        for i in range(idx, len(self._image_children)):
            self._image_children[i].idx = i

    def __contains__(self, item):
        if not (isinstance(item, AbstractKey)):
            return False
        if item == self:
            return True
        for image in self.image_children:
            if image == item:
                return True
        return any((item in folder for folder in self.folder_children))


class ImageKey(AbstractKey):
    def __init__(self, parent=None, idx: int = None, **kwargs):
        super().__init__(parent=parent, **kwargs)
        self.idx = idx

    @abstractmethod
    def get_image(self):
        pass

    def __contains__(self, item):
        if self == item:
            return True
        return False


class PathKey(AbstractKey):
    def __init__(self, parent, *, path: Path, **kwargs):
        self._path = path
        super().__init__(parent, **kwargs)

    @property
    def path(self):
        return self._path

    @property
    def name(self):
        return self._path.name

    def _file_key(self) -> str:
        return str(self._path.resolve())

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self._path == other._path
        else:
            return False

    def __repr__(self):
        return self.name


class H5Key(AbstractKey):
    def __init__(self, parent, *, h5path: Path, h5key: str = '', is_project: bool = None, **kwargs):
        super().__init__(parent, **kwargs)
        self._h5path: Path = h5path
        self._h5key: str = h5key
        self._is_project = is_project
        if self._is_project is None:
            self._is_project = _check_project(h5path)

    @property
    def h5path(self):
        return self._h5path

    @property
    def h5key(self):
        return self._h5key

    @property
    def is_project(self):
        return self._is_project

    @property
    def name(self):
        return self._h5key.split('/')[-1] if self._h5key else self._h5path.name

    def _file_key(self) -> str:
        return '-'.join((str(self._h5path.resolve()), self._h5key))

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return (self._h5path == other._h5path and
                    self._h5key == other._h5key)
        else:
            return False

    def __repr__(self):
        return self.name


class FolderH5Key(FolderKey, H5Key):
    def __init__(self, parent: FolderKey, *,
                 h5path: Path, h5key: str = '', is_project: bool = None):
        super().__init__(parent, h5path=h5path, h5key=h5key, is_project=is_project)

    def update(self):
        super().update()
        try:
            with File(str(self._h5path.resolve()), 'r') as f:
                if self._h5key:
                    f = f[self._h5key]
                for key in sorted(list(f.keys())):
                    item = f[key]
                    if isinstance(item, Group):
                        if self.is_project and IMAGE_PROJECT_KEY in item.attrs.keys():
                            self._image_children.append(
                                ImageH5Key(self, h5path=self._h5path,
                                           h5key='/'.join((self._h5key, key)), is_project=True,
                                           idx=len(self._image_children)))
                        else:
                            self._folder_children.append(
                                FolderH5Key(self, h5path=self._h5path,
                                            h5key='/'.join((self._h5key, key)),
                                            is_project=self.is_project))
                    elif isinstance(item, Dataset) and len(item.shape) == 2:
                        self._image_children.append(
                            ImageH5Key(self, h5path=self._h5path,
                                       h5key='/'.join((self._h5key, key)),
                                       is_project=False, idx=len(self._image_children)))
        except Exception as err:
            raise InvalidKey(err)

    def is_valid(self) -> bool:
        try:
            with File(str(self._h5path.resolve()), 'r') as f:
                if self._h5key:
                    f = f[self._h5key]
                if isinstance(f, Group):
                    return True
                else:
                    return False
        except (FileNotFoundError, KeyError, IOError):
            return False
        except Exception as err:
            logger.exception(err)
            return False


class FolderPathKey(FolderKey, PathKey):
    def __init__(self, parent: 'FolderKey', *, path: Path):
        super().__init__(parent, path=path)

    def update(self):
        super().update()
        try:
            for p in sorted(list(self._path.iterdir())):
                if p.is_dir():
                    self._folder_children.append(FolderPathKey(self, path=p))
                elif p.suffix in H5_FORMAT:
                    self._folder_children.append(FolderH5Key(self, h5path=p))
                elif p.suffix in AVAILABLE_IMAGE_FORMATS:
                    self._image_children.append(ImagePathKey(self, path=p, idx=len(self._image_children)))
        except Exception as err:
            raise InvalidKey(err)

    def is_valid(self) -> bool:
        return self._path.is_dir()


class ImagePathKey(ImageKey, PathKey):
    def __init__(self, parent: FolderPathKey, *,
                 path: Path, idx: int = None):
        super().__init__(parent, path=path, idx=idx)

    def get_image(self):
        try:
            return read_image(self._path)
        except Exception as err:
            logger.exception(err)
            return

    def is_valid(self) -> bool:
        return self._path.is_file()


class ImageH5Key(ImageKey, H5Key):
    def __init__(self, parent: FolderKey, *,
                 h5path: Path, h5key: str, is_project: bool = False, idx: int = None):
        super().__init__(
            parent, h5path=h5path, h5key=h5key, is_project=is_project, idx=idx)

    def get_image(self):
        try:
            with File(str(self._h5path.resolve()), 'r') as f:
                image = f[self._h5key]
                if self.is_project:
                    return image['image'][()]
                elif len(image.shape) == 2:
                    return image[()]
        except Exception as err:
            logger.exception(err)
            return

    def is_valid(self) -> bool:
        try:
            with File(str(self._h5path.resolve()), 'r') as f:
                dset = f[self._h5key]
                if not self.is_project and isinstance(dset, Dataset) and len(dset.shape) == 2:
                    return True
                elif self.is_project and isinstance(dset, Group):
                    return True
                return False
        except (FileNotFoundError, KeyError, IOError):
            return False
        except Exception as err:
            logger.exception(err)
            return False


class RemoveWeakrefs(object):
    def __init__(self, key: Union[FolderKey, ImageKey], *,
                 remove_subfolders: bool = False,
                 restore: bool = True):
        self.__key = key
        self.__parent = None
        self.__remove_subfolders: bool = remove_subfolders
        self.__restore: bool = restore
        self.__subfolders: List[FolderKey] = []

    def __enter__(self):
        if self.__remove_subfolders:
            try:
                self.__subfolders = self.__key.detach_subfolders()
            except AttributeError:
                self.__remove_subfolders = False
        self.__parent = self.remove(self.__key)
        return self.__key

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.__restore:
            self.restore(self.__key, self.__parent)

            if self.__remove_subfolders:
                self.__key.attach_subfolders(self.__subfolders)
        self.__subfolders = []
        self.__parent = None

    @staticmethod
    def restore(key: Union[FolderKey, ImageKey], parent: FolderKey = None) -> None:
        if parent:
            key.set_parent(parent)
        if isinstance(key, FolderKey):
            _restore_parents(key, parent)

    @staticmethod
    def remove(key: Union[FolderKey, ImageKey]) -> FolderKey or None:
        if isinstance(key, FolderKey):
            _remove_parents(key)
        return key.remove_parent()


def _remove_parents(key: FolderKey):
    for image_key in key.image_children:
        image_key.remove_parent()

    for folder_key in key.folder_children:
        folder_key.remove_parent()
        _remove_parents(folder_key)


def _restore_parents(key: FolderKey, parent: FolderKey = None):
    if parent:
        key.set_parent(parent)

    for image_key in key.image_children:
        image_key.set_parent(key)

    for folder_key in key.folder_children:
        folder_key.set_parent(folder_key)
        _restore_parents(folder_key)



from enum import Enum, auto

from PyQt5.QtWidgets import QWidget, QGridLayout
from PyQt5.QtCore import pyqtSignal

from ..app.update import CheckVersionMessage
from .basic_widgets import PopUpWrapper, TextNotification, Label, HyperlinkLabel


class NotificationTypes(Enum):
    checking_version = auto()
    updating_program = auto()
    check_result = auto()


def get_check_result_notification(res: CheckVersionMessage) -> TextNotification or None:
    if res.value == CheckVersionMessage.new_version_available.value:
        try:
            version = f'{res.version} '
        except AttributeError:
            return
            # version = ''
        return TextNotification('New version available',
                                f'The new version {version}will be installed')
    if res.value == CheckVersionMessage.latest_version_installed.value:
        return TextNotification('Latest version installed',
                                'You use the latest version of the program!')
    if res.value == CheckVersionMessage.error.value or res.value == CheckVersionMessage.no_internet.value:
        return TextNotification('Check version failed',
                                'Please, check your internet connection')
    if res.value == CheckVersionMessage.failed_updating.value:
        return TextNotification('Install failed', 'Please, check your internet connection or update manually')


class CheckingVersion(TextNotification):
    def __init__(self, parent=None):
        super().__init__('Check version', 'The program is looking for updates', spinner=True, parent=parent)


class UpdatingProgram(TextNotification):
    def __init__(self, version: str, parent=None):
        super().__init__('Install', f'Updating the program to version {version}', spinner=True, parent=parent)


class UpdateFailed(TextNotification):
    def __init__(self, parent=None):
        super().__init__('Install failed', 'Please, check your internet connection or use\n'
                                           'pip install --upgrade giwaxs_gui',
                         spinner=False, parent=parent)


class AskRestart(QWidget):
    sigRestart = pyqtSignal()

    def __init__(self, parent=None):
        super().__init__(parent)
        layout = QGridLayout(self)
        layout.setSpacing(5)
        title = Label('The program was updated', self, 8.5, True)
        title_font = title.font()
        title_font.setPointSizeF(8.5)
        title_font.setBold(True)
        title.setFont(title_font)
        layout.addWidget(title, 0, 0, 1, 2)
        layout.addWidget(Label('Please, restart the program to launch the new version.', self, 8.5), 1, 0, 1, 2)
        restart_label = HyperlinkLabel('Restart', self, 8.5)
        layout.addWidget(restart_label, 2, 0)
        restart_label.sigClicked.connect(self.sigRestart)



import logging
from typing import List
from datetime import datetime as dt

import numpy as np

from PyQt5.QtCore import pyqtSignal, pyqtSlot, QObject

from .rois.roi_dict import RoiDict, Roi
from .geometry import Geometry
from .geometry_holder import GeometryHolder
from .polar_image import (PolarImage, InterpolationParams,
                          INTERPOLATION_ALGORITHMS, INTERPOLATION_ALGORITHMS_INVERSED)
from .file_manager import FileManager, ImageKey
from .fitting import FitObject


class ImageHolder(QObject):
    sigImageChanged = pyqtSignal()
    sigPolarImageChanged = pyqtSignal()
    sigFitOpen = pyqtSignal(object)
    sigFitSaved = pyqtSignal(tuple)
    sigEmptyImage = pyqtSignal()

    log = logging.getLogger(__name__)

    def __init__(self, fm: FileManager, g_holder: GeometryHolder, roi_dict: RoiDict):
        QObject.__init__(self)
        self._fm = fm
        self._roi_dict = roi_dict
        self._image = self._raw_image = None
        self._current_key: ImageKey = None
        self._polar_image = PolarImage()
        self._g_holder = g_holder

        self._roi_dict.sigFitRoisOpen.connect(self.open_fit_rois)
        self._g_holder.sigPolarGeometryChanged.connect(self._update_polar_image)
        self._g_holder.sigGeometryChangeFinished.connect(self._update_polar_image)
        self._g_holder.sigTransformed.connect(self._update_image)

    @property
    def current_key(self) -> ImageKey or None:
        return self._current_key

    @property
    def image(self) -> np.ndarray or None:
        return self._image

    @property
    def raw_image(self) -> np.ndarray or None:
        return self._raw_image

    @property
    def polar(self) -> PolarImage:
        return self._polar_image

    @property
    def polar_image(self) -> np.ndarray or None:
        return self._polar_image.polar_image

    @property
    def polar_params(self) -> InterpolationParams:
        return self._polar_image.polar_params

    @property
    def g_holder(self) -> GeometryHolder:
        return self._g_holder

    @property
    def geometry(self) -> Geometry:
        return self.g_holder.geometry

    @pyqtSlot(object, name='changeImage')
    def change_image(self, image_key: ImageKey):
        if self._current_key == image_key:
            return
        self._current_key = image_key

        self._roi_dict.save_and_clear()

        if not image_key:
            self.g_holder.change_image(None)
            self._roi_dict.change_image(None)
            self.sigEmptyImage.emit()
            return

        image = self._fm.images[image_key]

        if image is None:
            self.sigEmptyImage.emit()
            return

        polar_image = self._fm.polar_images[image_key]
        prev_geometry = self.geometry

        self._raw_image = image
        self._image = self.g_holder.change_image(image_key, image)
        self._update_polar_image(polar_image, False)

        if self.geometry.beam_center != prev_geometry.beam_center:
            self.g_holder.sigBeamCenterChanged.emit()
        if self.geometry.scale != prev_geometry.scale:
            self.g_holder.sigScaleChanged.emit()

        self.sigImageChanged.emit()
        self.sigPolarImageChanged.emit()

        # self.g_holder.check_ring_bounds()
        self._roi_dict.change_image(image_key)

    def get_data_by_key(self, image_key: ImageKey, save: bool = False):
        polar_image = self._fm.polar_images[image_key]
        image = self._fm.images[image_key]
        geometry = self.g_holder.get_geometry(image_key)
        if image is None or geometry is None:
            return None, None, None

        image = geometry.t(image)
        if geometry.shape != image.shape:
            geometry.set_shape(image.shape)
        if polar_image is None:
            yy, zz = geometry.polar_grids
            polar_image = self.polar.calc_polar_image(image, yy, zz, self.polar_params.algorithm)
            if save:
                self._fm.polar_images[image_key] = polar_image
        return image, polar_image, geometry

    # def set_image(self, img: np.ndarray, polar_image: np.ndarray = None):
    #     self._raw_image = img
    #     self._update_image()
    #     self.update_polar_image(polar_image)

    def _update_polar_image(self, polar_image=None, emit: bool = True):
        if polar_image is not None:
            self._polar_image.set_polar_image(polar_image)
        else:
            self._polar_image.update(self.geometry, self.image)
        if emit:
            self.sigPolarImageChanged.emit()

    def _update_image(self, polar_image: np.ndarray = None, emit: bool = True) -> None:
        if self.raw_image is None:
            return
        old_shape = self.image.shape if self.image is not None else None
        self._image = self.g_holder.transform_image(self._raw_image)
        if old_shape != self.image.shape:
            self.g_holder.set_shape(self.image.shape)
        self._update_polar_image(polar_image, emit=emit)
        if emit:
            self.sigImageChanged.emit()

    def set_polar_image_params(self, params: dict):
        shape = (params.get('phi_size', self.polar_params.shape[0]),
                 params.get('r_size', self.polar_params.shape[1]))
        algorithm = INTERPOLATION_ALGORITHMS.get(params.get('mode', None), self.polar_params.algorithm)
        self.polar.set_params(shape=shape, algorithm=algorithm)
        self.g_holder.set_polar_shape(shape)

    def polar_image_params_dict(self):
        return dict(phi_size=self.geometry.polar_shape[0], r_size=self.geometry.polar_shape[1],
                    algorithm=INTERPOLATION_ALGORITHMS_INVERSED[self.polar_params.algorithm])

    def get_radial_profile(self) -> np.ndarray or None:
        return self.polar.get_radial_profile()

    def get_angular_profile(self, key: int) -> np.ndarray or None:
        if key is None:
            try:
                roi = self._roi_dict.selected_rois[0]
            except IndexError:
                return
        else:
            roi = self._roi_dict[key]

        return self.polar.get_angular_profile(self.geometry, roi)

    @pyqtSlot(list, name='openFitRois')
    def open_fit_rois(self, rois: List[Roi]):
        fit_object = FitObject(self._current_key, self.polar_image,
                               self.geometry.r_axis, self.geometry.phi_axis)
        profile = self._fm.profiles[self._current_key]

        if profile:
            fit_object.set_profile(profile)

        for roi in rois:
            fit_object.new_fit(roi)

        self.sigFitOpen.emit(fit_object)

    @pyqtSlot(object, name='applyFit')
    def apply_fit(self, fit_object: FitObject):
        name = dt.now().ctime()
        fit_object.name = name
        parent = fit_object.image_key.parent
        fit_object.image_key.remove_parent()
        self._fm.fits[fit_object.image_key, name] = fit_object
        fit_object.image_key.set_parent(parent)

        self._roi_dict.apply_fit([fit.roi for fit in fit_object.fits.values()], fit_object.image_key)

        self.sigFitSaved.emit((fit_object.image_key, name))



from dataclasses import dataclass
from enum import Enum

import numpy as np


class RoiTypes(Enum):
    ring = 1
    segment = 2


@dataclass
class Roi:
    radius: float
    width: float
    angle: float = 180
    angle_std: float = 360
    key: int = None
    name: str = ''
    group: str = ''
    type: RoiTypes = RoiTypes.ring
    movable: bool = True
    fitted_parameters: dict = None
    active: bool = False

    def update(self, other: 'Roi'):
        self.__dict__ = other.__dict__

    def to_array(self) -> np.ndarray:
        return np.array([self.radius, self.width, self.angle, self.angle_std,
                         self.key, self.type.value])

    @classmethod
    def from_array(cls, arr: np.ndarray, **meta_data):
        roi = cls(**dict(zip(_ROI_NAMES, arr)), **meta_data)
        roi.type = RoiTypes(roi.type)
        return roi


DTYPES = [('radius', 'f4'), ('width', 'f4'), ('angle', 'f4'), ('angle_std', 'f4'),
          ('key', 'i4'), ('type', 'i4')]

_ROI_NAMES = list(map(lambda x: x[0], DTYPES))


def roi_to_tuple(roi: Roi):
    return (roi.radius, roi.width, roi.angle, roi.angle_std,
            roi.key, roi.name, roi.group, roi.type.value)



from typing import Tuple
import logging

from PyQt5.QtCore import pyqtSignal, pyqtSlot, QObject

import numpy as np

from .transformations import Transformation
from .file_manager import FileManager, ImageKey
from .geometry import Geometry


class GeometryHolder(QObject):
    sigBeamCenterChanged = pyqtSignal()
    sigGeometryChangeFinished = pyqtSignal()
    sigPolarGeometryChanged = pyqtSignal()
    sigTransformed = pyqtSignal()
    sigScaleChanged = pyqtSignal()
    sigRingBoundsChanged = pyqtSignal(tuple)

    log = logging.getLogger(__name__)

    def __init__(self, fm: FileManager):
        QObject.__init__(self)
        self._fm = fm
        self._default_geometry = Geometry()
        self._current_geometry = None
        self._ring_bounds = (0, np.pi * 2)
        self._current_key: ImageKey or None = None

    def transform_image(self, raw_image: np.ndarray):
        return self.geometry.t(raw_image)

    def add_transform(self, op: Transformation):
        if not self._current_geometry:
            self._current_geometry = self.geometry.copy()
        self.geometry.t.add(op)
        self.sigTransformed.emit()

    def get_geometry(self, image_key: ImageKey):
        return self._fm.geometries[image_key] or self._fm.geometries.default[image_key] or Geometry()

    @property
    def geometry(self) -> Geometry:
        return self._current_geometry or self._default_geometry

    @property
    def is_default(self) -> bool:
        return self._current_geometry is None

    def change_image(self, image_key: ImageKey, image: np.ndarray = None):
        if image_key == self._current_key:
            return

        self.save_state()

        self._current_key = image_key
        if not self._current_key:
            return

        self._default_geometry = self._fm.geometries.default[image_key] or Geometry()
        self._current_geometry = self._fm.geometries[image_key]

        if image is not None:
            image = self.transform_image(image)
            if image.shape != self.geometry.shape:
                self.set_shape(image.shape)
        return image

    def save_as_default(self):
        if self._current_geometry and self._current_key:
            self._default_geometry = self._current_geometry
            self._fm.geometries.default[self._current_key] = self._current_geometry
            del self._fm.geometries[self._current_key]
            self._current_geometry = None
            self.log.info('Geometry saved as default.')

        self.log.info('Geometry not saved (default used)')

    def save_state(self):
        if not (self._current_key and self._current_geometry):
            self.log.info('Geometry not saved (default used)')
            return

        if not self._fm.geometries.default[self._current_key]:
            self.save_as_default()
        elif self._current_geometry != self._default_geometry:
            self._fm.geometries[self._current_key] = self._current_geometry
            self.log.info('Non-default geometry saved')
        else:
            del self._fm.geometries[self._current_key]
            self.log.info('Non-default geometry deleted.')

    @pyqtSlot(tuple, bool, name='changeBeamCenter')
    def set_beam_center(self, beam_center: tuple, finished: bool = True):
        if (beam_center[0] == self.geometry.beam_center.z and
                beam_center[1] == self.geometry.beam_center.y):
            return
        if not self._current_geometry:
            self._current_geometry = self.geometry.copy()
        self._current_geometry.set_beam_center(*beam_center)
        self.sigBeamCenterChanged.emit()
        self.check_ring_bounds()
        if finished:
            self.sigGeometryChangeFinished.emit()

    def set_shape(self, shape: Tuple[int, int]):
        if not self._current_geometry:
            self._current_geometry = self.geometry.copy()
        self._current_geometry.set_shape(shape)

    @pyqtSlot(tuple, name='changePolarShape')
    def set_polar_shape(self, shape: Tuple[int, int]):
        if not self._current_geometry:
            self._current_geometry = self.geometry.copy()
        self._current_geometry.set_polar_shape(shape)
        self.sigGeometryChangeFinished.emit()

    @pyqtSlot(float, name='changeScale')
    def set_scale(self, scale: float):
        if self.geometry.scale == scale:
            return
        if not self._current_geometry:
            self._current_geometry = self.geometry.copy()
        self._current_geometry.set_scale(scale)
        self.sigScaleChanged.emit()

    def check_ring_bounds(self):
        if self.geometry and self.geometry.ring_bounds != self._ring_bounds:
            self._ring_bounds = self.geometry.ring_bounds
            self.sigRingBoundsChanged.emit(self._ring_bounds)



from abc import abstractmethod
from typing import Callable

from PyQt5.QtGui import QColor

from .rois_context_menu import RoiContextMenu
from ...app.rois.roi import Roi

COLOR_DICT = dict(
    default=QColor(0, 0, 255),
    active=QColor(0, 128, 255),
    fixed=QColor(0, 255, 0),
    fixed_active=QColor(255, 0, 255)
)


def _color_key(roi: Roi) -> str:
    if roi.active and roi.movable:
        return 'active'
    if not roi.active and roi.movable:
        return 'default'
    if roi.active and not roi.movable:
        return 'fixed_active'
    return 'fixed'


class AbstractRoiWidget(object):
    BRIGHT_COLOR: bool = True

    def __init__(self, roi: Roi, enable_context: bool = False, context_menu: Callable[[Roi], None] = RoiContextMenu):
        self._roi = roi
        if enable_context:
            self._context_menu = context_menu
        else:
            self._context_menu = None

    def update_color(self):
        color = COLOR_DICT[_color_key(self.roi)]
        if not self.BRIGHT_COLOR:
            color.setAlpha(150)
        self.set_color(color)

    def update_roi(self):

        self.move_roi()

        if not self.roi.movable:
            self.fix()
        else:
            self.unfix()
        self.update_color()

    @abstractmethod
    def move_roi(self):
        pass

    @abstractmethod
    def send_move(self):
        pass

    @property
    def roi(self):
        return self._roi

    def set_roi(self, roi: Roi, update: bool = True):
        self._roi = roi
        if update:
            if not self.roi.movable:
                self.fix()
            self.update_color()
            self.move_roi()

    @abstractmethod
    def set_color(self, color):
        pass

    def show_context_menu(self, *args):
        if self._context_menu:
            self._context_menu(self.roi)

    def update_select(self):
        self.update_color()

    def fix(self):
        self.update_color()

    def unfix(self):
        self.update_color()

    # def show_roi(self):
    #     self.show()
    #
    # def hide_roi(self):
    #     self.hide()



import json
from enum import Enum, auto
import logging
import requests
from pkg_resources import parse_version

logger = logging.getLogger(__name__)


class CheckVersionMessage(Enum):
    new_version_available = auto()
    latest_version_installed = auto()
    no_internet = auto()
    error = auto()
    not_checked = auto()
    updated = auto()
    failed_updating = auto()


def check_outdated(version: str, package: str = 'giwaxs_gui') -> CheckVersionMessage:
    logger.info(f'Checking the latest version of the {package} package.')

    url = f'https://pypi.python.org/pypi/{package}/json'

    try:
        response = requests.get(url).text
    except requests.ConnectionError:
        logger.info(f'No internet.')
        return CheckVersionMessage.no_internet

    try:
        current_version = parse_version(version)
        latest_version = json.loads(response)['info']['version']
        latest_version = parse_version(latest_version)

        if current_version < latest_version:
            message = CheckVersionMessage.new_version_available
            message.version = latest_version
            logger.info(f'Current version: {current_version}')
            logger.info(f'New version available : {latest_version}.')
            return message
        else:
            message = CheckVersionMessage.latest_version_installed
            message.version = latest_version
            logger.info(f'latest_version_installed : {message.version}.')
            return message

    except Exception as err:
        logger.exception(err)
        return CheckVersionMessage.error



import sys
import argparse
import logging

from PyQt5.QtWidgets import QApplication
from PyQt5.QtCore import Qt

from giwaxs_gui.app import App, launch_detached
from giwaxs_gui.gui import GIWAXSMainController, UncaughtHook, DebugWindow
from giwaxs_gui.__version import __version__


__author__ = 'Vladimir Starostin'
__email__ = 'v.starostin.m@gmail.com'


def run(logging_level: int = logging.ERROR):
    for log in (logging.getLogger(name) for name in logging.root.manager.loggerDict):
        log.setLevel(logging_level)

    QApplication.setAttribute(Qt.AA_EnableHighDpiScaling, True)
    q_app = QApplication([])

    giwaxs_app = GIWAXSMainController()

    return q_app.exec_()


def main():
    parser = argparse.ArgumentParser(description='Program for GIWAXS analysis')

    parser.add_argument('-d', '--debug', action='store_true', help='open program in debug mode')
    parser.add_argument('--version', action='store_true', help='show the installed version')

    args = parser.parse_args()

    if args.version:
        print(__version__)
        return

    level: int = logging.DEBUG if args.debug else logging.ERROR

    exit_code: int = run(level)

    if exit_code == GIWAXSMainController.EXIT_CODE_REBOOT:
        launch_detached()
        return 0

    return exit_code


if __name__ == '__main__':
    sys.exit(main())



import logging

from .rois.roi_dict import RoiDict
from .file_manager import FileManager
from .utils import SingletonMeta
from .geometry import Geometry
from .geometry_holder import GeometryHolder
from .image_holder import ImageHolder
from .profiles import RadialProfile, AngularProfile
from .debug_tracker import TrackQObjects


class App(metaclass=SingletonMeta):
    log = logging.getLogger(__name__)

    def __init__(self):
        self.fm: FileManager = FileManager()
        self.geometry_holder: GeometryHolder = GeometryHolder(self.fm)
        self.roi_dict: RoiDict = RoiDict(self.fm, self.geometry_holder)
        self.image_holder: ImageHolder = ImageHolder(self.fm, self.geometry_holder, self.roi_dict)
        self.radial_profile: RadialProfile = RadialProfile(self.image_holder, self.fm)
        self.angular_profile: AngularProfile = AngularProfile(self.image_holder)

        self.image_holder.sigPolarImageChanged.connect(self.radial_profile.update)
        self.image_holder.sigPolarImageChanged.connect(self.angular_profile.update)

        self.geometry_holder.sigScaleChanged.connect(self.roi_dict.on_scale_changed)
        self.geometry_holder.sigRingBoundsChanged.connect(self.roi_dict.change_ring_bounds)

        self.fm.sigActiveImageChanged.connect(self.image_holder.change_image)
        self.fm.sigProjectIsClosing.connect(self.save_state)

        if self.log.level <= logging.DEBUG:
            self.debug_tracker: TrackQObjects = TrackQObjects()
        else:
            self.debug_tracker: TrackQObjects = None

    def save_state(self):
        self.geometry_holder.save_state()
        self.roi_dict.save_state()
        # self.radial_profile.save_state()

    @property
    def image(self):
        return self.image_holder.image

    @property
    def polar_image(self):
        return self.image_holder.polar_image

    @property
    def geometry(self) -> Geometry:
        return self.geometry_holder.geometry

    def close(self):
        self.fm.close()



from PyQt5.QtWidgets import QLabel
from PyQt5.QtCore import pyqtSignal, Qt


class Label(QLabel):
    def __init__(self, text: str, parent=None, font_size: float = 8., bold: bool = False):
        super().__init__(text, parent=parent)
        font = self.font()
        font.setPointSizeF(font_size)
        font.setBold(bold)
        self.setFont(font)


class LabelButton(Label):
    sigClicked = pyqtSignal()
    sigPressed = pyqtSignal()

    def mousePressEvent(self, ev) -> None:
        self.sigPressed.emit()
        ev.accept()

    def mouseReleaseEvent(self, ev) -> None:
        self.sigClicked.emit()
        ev.accept()

    def enterEvent(self, a0) -> None:
        self.setStyleSheet('QLabel {color : #D3D3D3}')
        super().enterEvent(a0)

    def leaveEvent(self, a0) -> None:
        self.setStyleSheet('QLabel {color : #808080}')
        super().leaveEvent(a0)


class HyperlinkLabel(LabelButton):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setStyleSheet('QLabel {color : #71bbd4}')
        self.setAttribute(Qt.WA_TranslucentBackground, True)

    def enterEvent(self, a0) -> None:
        font = self.font()
        font.setUnderline(True)
        self.setFont(font)
        QLabel.enterEvent(self, a0)

    def leaveEvent(self, a0) -> None:
        font = self.font()
        font.setUnderline(False)
        self.setFont(font)
        QLabel.leaveEvent(self, a0)



import logging
from enum import Enum
from typing import Dict, List, Tuple
from time import sleep
from copy import deepcopy

from PyQt5.QtCore import (QObject, pyqtSlot, pyqtSignal,
                          QCoreApplication, Qt, QThread)
from PyQt5.QtWidgets import (QWidget, QVBoxLayout, QPushButton,
                             QProgressBar, QSlider, QLabel, QMessageBox)

from pyqtgraph import GraphicsLayoutWidget, FillBetweenItem, InfiniteLine

from ...app import App, Roi, RoiData
from ...app.file_manager import ImageKey, FolderKey
from ...app.fitting import FitObject, Fit

from ..tools import get_pen, center_widget, Icon, show_error

logger = logging.getLogger(__name__)


class ButtonStates(Enum):
    start = 'Start fit'
    pause = 'Pause fit'
    resume = 'Resume fit here'
    finished = 'Fit is finished'


class MultiFit(QObject):
    sigFit = pyqtSignal(object)
    sigPaused = pyqtSignal()
    # sigError = pyqtSignal(object)
    sigFinished = pyqtSignal()
    sigSaved = pyqtSignal(int)
    sigSavedFinished = pyqtSignal()
    sigDeleted = pyqtSignal(int)
    sigDeletedFinished = pyqtSignal()

    log = logging.getLogger(__name__)

    def __init__(self, fm_multi_fit, folder_key: FolderKey, parent=None):
        super().__init__(parent=parent)
        self.sleep_time: float = 0.002
        self._paused: bool = True
        self._stopped: bool = False
        self.fm_multi_fit = fm_multi_fit
        self.folder_key: FolderKey = folder_key

    @pyqtSlot(list, name='runSave')
    def run_save(self, image_key_list: list):
        rois_data_fm = App().fm.rois_data

        for i, image_key in enumerate(image_key_list):

            fit_obj = self.fm_multi_fit[image_key]
            if not fit_obj:
                self.sigSaved.emit(i)
                continue

            rois = [fit.roi for fit in fit_obj.fits.values() if fit.fitted_params]

            if rois:
                roi_data = rois_data_fm[image_key] or RoiData()
                roi_data.apply_fit(rois)
                rois_data_fm[image_key] = roi_data

            self.sigSaved.emit(i)

            if self.sleep_time:
                sleep(self.sleep_time)

        self.sigSavedFinished.emit()

    @pyqtSlot(object, name='runFit')
    def run_fit(self, fit_obj: FitObject):
        self._paused = False
        fit_obj = deepcopy(fit_obj)

        while fit_obj and fit_obj.fits and not self._paused:

            self.log.info(f'Fitting image {fit_obj.image_key}')
            for fit in fit_obj.fits.values():
                if fit.fitted_params:
                    continue
                try:
                    fit.do_fit()
                except Exception as err:
                    self.log.exception(err)

                fit.roi.movable = True

                if self.sleep_time:
                    sleep(self.sleep_time)

                QCoreApplication.processEvents()

                if self._paused:
                    self.log.debug('Paused!')
                    break

            if self._paused:
                break

            fit_obj.is_fitted = True
            new_key = fit_obj.image_key.parent.get_next_image(fit_obj.image_key)

            if not new_key:
                self.sigFit.emit(fit_obj)
                break

            saved_fit = self.fm_multi_fit[new_key]
            self.sigFit.emit(deepcopy(fit_obj))
            new_fit_obj = get_new_fit(fit_obj, saved_fit=saved_fit,
                                      add_fits=True, new_image_key=new_key)
            fit_obj = new_fit_obj

        if self._stopped:
            self.deleteLater()
        elif self._paused:
            self.sigPaused.emit()
        else:
            self._paused = True
            self.sigFinished.emit()

    @pyqtSlot(list, int, name='runFit')
    def run_delete(self, image_list: List[ImageKey], roi_key: int):
        for i, image_key in enumerate(self.folder_key.image_children):
            if image_key in image_list:
                fit_obj = self.fm_multi_fit[image_key]
                if fit_obj:
                    try:
                        del fit_obj.fits[roi_key]
                        self.fm_multi_fit[image_key] = fit_obj
                    except KeyError:
                        pass
            self.sigDeleted.emit(i)
            if self.sleep_time:
                sleep(self.sleep_time)
        self.sigDeletedFinished.emit()

    @property
    def is_paused(self):
        return self._paused

    @pyqtSlot(name='pauseFit')
    def pause(self):
        self.log.debug('Pause signal received')
        self._paused = True

    @pyqtSlot(name='stopFit')
    def stop(self):
        self.log.debug('Stop signal received')
        self._paused = True
        self._stopped = True


def get_new_fit(previous_fit: FitObject, saved_fit: FitObject = None,
                add_fits: bool = False, new_image_key: ImageKey = None):
    if not saved_fit:
        if not new_image_key:
            folder_key: FolderKey = previous_fit.image_key.parent
            if not folder_key:
                logger.debug('empty folder key')
                return
            new_image_key: ImageKey = folder_key.get_next_image(previous_fit.image_key)
        if not new_image_key:
            logger.debug('empty next image_key')
            return

        image, polar_image, geometry = App().image_holder.get_data_by_key(new_image_key, save=True)

        if polar_image is None:
            logger.debug('empty polar_image')
            return

        saved_fit: FitObject = FitObject(new_image_key, polar_image, geometry.r_axis, geometry.phi_axis)

    if not saved_fit.saved_profile:

        saved_profile = App().fm.profiles[new_image_key]

        if saved_profile:
            saved_fit.set_profile(saved_profile, update_baseline=False)
        elif previous_fit.saved_profile:
            saved_fit.set_profile(previous_fit.saved_profile, update_baseline=True)
            App().fm.profiles[new_image_key] = saved_fit.saved_profile

    if add_fits:
        for fit in previous_fit.fits.values():
            if fit.fitted_params and fit.roi.key not in saved_fit.fits.keys():
                saved_fit.add_fit(fit)
                fit.fitted_params = None

    return saved_fit


class MultiFitWindow(QWidget):
    sigPauseFit = pyqtSignal()
    sigDeleteFit = pyqtSignal()
    sigRunFit = pyqtSignal(object)
    sigClosed = pyqtSignal()
    sigFitUpdated = pyqtSignal(object)
    sigRunSave = pyqtSignal(list)
    sigRunDelete = pyqtSignal(list, int)

    log = logging.getLogger(__name__)

    def __init__(self, fit_object: FitObject, parent=None):
        super().__init__(parent=parent)
        self.current_fit: FitObject = fit_object
        self.folder_key: FolderKey = self.current_fit.image_key.parent
        self.fm = App().fm.fits.get_multi_fit()

        self.fit_thread = QThread()
        self.multi_fit = MultiFit(self.fm, self.folder_key)
        self.multi_fit.moveToThread(self.fit_thread)

        self.multi_fit.sigPaused.connect(self.on_paused)
        self.multi_fit.sigFinished.connect(self.on_finished)
        self.multi_fit.sigFit.connect(self._update_fit)

        self.sigRunFit.connect(self.multi_fit.run_fit)
        self.sigPauseFit.connect(self.multi_fit.pause)
        self.sigDeleteFit.connect(self.multi_fit.stop)
        self.sigRunSave.connect(self.multi_fit.run_save)
        self.sigRunDelete.connect(self.multi_fit.run_delete)

        self.fit_thread.finished.connect(self.multi_fit.deleteLater)
        self.fit_thread.finished.connect(self.fit_thread.deleteLater)

        self._init_ui()
        self.fit_thread.start()

        self.plot_params.add_fit(fit_object)

        if App().debug_tracker:
            App().debug_tracker.add_object(self)
            App().debug_tracker.add_object(self.multi_fit)
            App().debug_tracker.add_object(self.fit_thread)

        self.show()

    def _init_ui(self):
        layout = QVBoxLayout(self)
        self.plot_params = MultiFitPlot(self)
        self.progress_widget = ImageSeriesSliderProgressWidget(self.current_fit.image_key, self)
        self.control_button = QPushButton(ButtonStates.start.value)
        layout.addWidget(QLabel('Image series'))
        layout.addWidget(self.plot_params)
        layout.addWidget(self.progress_widget)
        layout.addWidget(self.control_button)

        self.control_button.clicked.connect(self._on_button_clicked)
        self.progress_widget.sigImageChanged.connect(self._change_image)

    @pyqtSlot(name='controlButtonClicked')
    def _on_button_clicked(self):
        self.log.debug('Button clicked')
        if self.control_button.text() == ButtonStates.start.value:
            self._start_fit()
        elif self.control_button.text() == ButtonStates.pause.value:
            self.log.debug('Emitting pause signal')
            self.sigPauseFit.emit()
        elif self.control_button.text() == ButtonStates.resume.value:
            self._start_fit()

    @pyqtSlot(object, name='changeImage')
    def _change_image(self, key: ImageKey):
        if key == self.current_fit.image_key:
            return

        add_fits: bool = self.current_fit.image_key.idx + 1 == key.idx if self.current_fit else False

        self.save_current_fit()
        self.current_fit = get_new_fit(
            self.current_fit, saved_fit=self.fm[key], new_image_key=key, add_fits=add_fits)
        self.plot_params.change_image(key)

        if (self.control_button.text() == ButtonStates.finished.value and
                key.idx + 1 < key.parent.images_num):
            self.control_button.setText(ButtonStates.resume.value)

        self.sigFitUpdated.emit(self.current_fit)

    def save_current_fit(self):
        if self.current_fit:
            self.fm[self.current_fit.image_key] = self.current_fit

    def update_fit(self, fit: Fit):
        if self.current_fit:
            self.plot_params.update_roi(fit, self.current_fit.image_key.idx)

    def save_fits(self, image_keys: list):
        self.log.info(f'Saving {len(image_keys)} fits ...')

        if not self.multi_fit.is_paused:
            show_error('Please, stop the auto fitting process before saving the images',
                       error_title='Saving fits failed')
            return

        widget = ProgressWidget(len(image_keys),
                                progress_label='Applying fits ...',
                                finish_label='Fits are saved!', parent=self)

        if image_keys:
            self.multi_fit.sigSaved.connect(widget.set_progress)
            self.multi_fit.sigSavedFinished.connect(widget.finished)
            self.sigRunSave.emit(image_keys)

    def delete_rois(self, image_keys: List[ImageKey], roi_key: int):
        self.log.info(f'Deleting roi {roi_key} from {len(image_keys)} images')

        if len(image_keys) == 1 and self.current_fit and image_keys[0] == self.current_fit.image_key:
            self.plot_params.delete_roi(roi_key, self.current_fit.image_key.idx)
            return

        if not self.multi_fit.is_paused:
            show_error('Please, stop the auto fitting process before deleting rois',
                       error_title='Delete failed')
            return

        widget = ProgressWidget(self.folder_key.images_num,
                                progress_label='Deleting roi ...',
                                finish_label='Roi is deleted!', parent=self)

        for image_key in image_keys:
            self.plot_params.delete_roi(roi_key, image_key.idx)

        self.multi_fit.sigDeleted.connect(widget.set_progress)
        self.multi_fit.sigDeletedFinished.connect(widget.finished)
        self.sigRunDelete.emit(image_keys, roi_key)

    @pyqtSlot(name='fitPaused')
    def on_paused(self):
        self.control_button.setText(ButtonStates.resume.value)
        self.progress_widget.set_fixed(False)

    def _start_fit(self):
        self.control_button.setText(ButtonStates.pause.value)
        self.progress_widget.set_fixed(True)
        self.sigRunFit.emit(self.current_fit)

    @pyqtSlot(name='fitFinished')
    def on_finished(self):
        self.control_button.setText(ButtonStates.finished.value)
        self.progress_widget.set_fixed(False)
        # self.control_button.setDisabled(True)

    @pyqtSlot(object, name='updateFit')
    def _update_fit(self, fit_object: FitObject or None):
        self.log.debug('Updating fit ...')
        self.plot_params.add_fit(fit_object)
        self.current_fit = fit_object

        self.save_current_fit()

        self.progress_widget.change_image(fit_object.image_key)
        self.sigFitUpdated.emit(fit_object)

    def select_fit(self, key: int):
        self.plot_params.select_fit(key)

    def unselect_fit(self, key: int):
        self.plot_params.unselect_fit(key)

    def close_widget(self) -> None:
        self.log.debug(f'Closing multi fit object...')
        self.fm.delete()
        self.sigDeleteFit.emit()
        if self.fit_thread.isRunning():
            self.fit_thread.quit()
            self.fit_thread.wait()

        self.sigClosed.emit()
        self.deleteLater()


class MultiFitPlot(GraphicsLayoutWidget):
    log = logging.getLogger(__name__)

    INACTIVE_COLOR = [63, 63, 63, 190]
    ACTIVE_COLOR = [50, 250, 50, 240]

    def __init__(self, parent=None):
        super().__init__(parent)
        self.plot_item = self.addPlot()
        self.plot_item.setMenuEnabled(False)
        self.inf_line = InfiniteLine(0, pen=get_pen(color='red', style=Qt.DashLine))
        self.plot_item.addItem(self.inf_line)
        self.plots = {}
        self.x_axis: Dict[int, List[int]] = {}

    def add_fit(self, fit_obj: FitObject):
        self.log.debug(f'Number of fits = {len(fit_obj.fits)}')
        x = fit_obj.image_key.idx

        for fit in fit_obj.fits.values():
            self._add_fit(fit, x)

        fitted_keys = [k for k in fit_obj.fits.keys() if fit_obj.fits[k].fitted_params]

        for key, x_axis in self.x_axis.items():
            if x in x_axis and (key not in fitted_keys):
                self._delete_fit(key, x)

        self.change_image(fit_obj.image_key)

    def update_roi(self, fit: Fit, x: int):
        self._add_fit(fit, x)

    def delete_roi(self, roi_key: int, image_idx: int):
        self._delete_fit(roi_key, image_idx)

    def _add_fit(self, fit: Fit, x):
        # if not fit.roi.fitted_parameters:
        #     return

        key = fit.roi.key

        if key not in self.plots:
            self._init_plot(key)

        idx, x_axis = self._add_x(key, x)

        plots = self.plots[key]

        radius, width = fit.roi.radius, fit.roi.width

        for name, point in zip(('upper', 'middle', 'lower'),
                               (radius + width, radius, radius - width)):
            plot = plots[name]
            y = plot.yData
            if y is None:
                y = [point]
            else:
                y = y.tolist()
                if len(y) == len(x_axis):
                    # self.log.debug('Moving roi ...')
                    y[idx] = point
                else:
                    # self.log.debug('Inserting roi ...')
                    y.insert(idx, point)
            plot.setData(x_axis, y)

        self._update_fill_between(plots)

    def _add_x(self, key: int, x) -> Tuple[int, list]:
        if key not in self.x_axis:
            self.x_axis[key] = [x]
            return 0, self.x_axis[key]

        x_axis: list = self.x_axis[key]
        if x not in x_axis:
            if not len(x_axis) or x_axis[-1] < x:
                x_axis.append(x)
                return len(x_axis) - 1, x_axis
            else:
                # TODO implement binary search
                for i, other_x in enumerate(x_axis):
                    if x < other_x:
                        x_axis.insert(i, x)
                        return i, x_axis
        else:
            return x_axis.index(x), x_axis

    def _delete_fit(self, key: int, x: int):
        try:
            plots = self.plots[key]
            x_axis = self.x_axis[key]
            idx = x_axis.index(x)
        except (KeyError, ValueError):
            return

        # self.log.debug('Deleting fit ... ')

        x_axis.remove(x)

        for name in ('upper', 'middle', 'lower'):
            plot = plots[name]
            y = plot.yData.tolist()
            del y[idx]
            plot.setData(x_axis, y)

        self._update_fill_between(plots)

    def select_fit(self, key: int):
        try:
            self.plots[key]['fill'].setBrush(self.ACTIVE_COLOR)
        except KeyError:
            pass

    def unselect_fit(self, key: int):
        try:
            self.plots[key]['fill'].setBrush(self.INACTIVE_COLOR)
        except KeyError:
            pass

    def _init_plot(self, key):
        self.plots[key] = {}
        self.plots[key]['upper'] = self.plot_item.plot(pen=get_pen(color='blue'))
        self.plots[key]['middle'] = self.plot_item.plot(
            pen=get_pen(style=Qt.DashLine)
        )
        self.plots[key]['lower'] = self.plot_item.plot(pen=get_pen(color='blue'))

    def _update_fill_between(self, plots):
        if 'fill' not in plots:
            plots['fill'] = FillBetweenItem(plots['upper'], plots['lower'],
                                            brush=self.INACTIVE_COLOR)
            self.plot_item.addItem(plots['fill'])
        else:
            plots['fill'].setCurves(plots['upper'], plots['lower'])

    @pyqtSlot(object, name='changeImage')
    def change_image(self, key: ImageKey):
        self.inf_line.setValue(key.idx)
        self.plot_item.autoRange()


class ImageSeriesSliderProgressWidget(QWidget):
    sigImageChanged = pyqtSignal(object)

    def __init__(self, image_key: ImageKey, parent=None):
        super().__init__(parent)
        self.current_key: ImageKey = image_key
        self.folder_key: FolderKey = image_key.parent
        self._init_ui()
        self.set_fixed(False)

    def _init_ui(self):
        layout = QVBoxLayout(self)
        self.progress_bar = QProgressBar(self)
        self.slider = QSlider(orientation=Qt.Horizontal, parent=self)
        self.label = QLabel(self._label_text())

        self.progress_bar.setMaximum(self.folder_key.images_num - 1)
        self.slider.setMaximum(self.folder_key.images_num - 1)
        self.slider.valueChanged.connect(self._on_slider_moved)
        if self.current_key:
            self.slider.setValue(self.current_key.idx)

        layout.addWidget(self.label)
        layout.addWidget(self.progress_bar)
        layout.addWidget(self.slider)

    @pyqtSlot(bool, name='setFixed')
    def set_fixed(self, fixed: bool):
        if fixed:
            self.progress_bar.show()
            self.slider.hide()
        else:
            self.progress_bar.hide()
            self.slider.show()

    @pyqtSlot(int, name='sliderMoved')
    def _on_slider_moved(self, value: int):
        image_key = self.folder_key.image_by_key(value)
        if image_key and image_key != self.current_key:
            self.current_key = image_key
            self.sigImageChanged.emit(image_key)
            self.label.setText(self._label_text())

    @pyqtSlot(object, name='changeImage')
    def change_image(self, image_key: ImageKey):
        self.current_key = image_key
        self.progress_bar.setValue(image_key.idx)
        self.label.setText(self._label_text())
        self.slider.setValue(image_key.idx)

    def _label_text(self):
        return f'Image {self.current_key.idx}: {self.current_key.name}'


class ProgressWidget(QWidget):
    def __init__(self, num: int, progress_label: str, finish_label: str, parent=None):
        super().__init__(parent)
        self.progress_label: str = progress_label
        self.finish_label: str = finish_label

        self.setWindowFlag(Qt.Window)
        self.setWindowFlag(Qt.FramelessWindowHint)
        self.setWindowModality(Qt.ApplicationModal)
        self.setAttribute(Qt.WA_DeleteOnClose, True)

        self._init_ui(num)
        center_widget(self)
        self.show()

    def _init_ui(self, num: int):
        layout = QVBoxLayout(self)
        self.progress = QProgressBar()
        self.progress.setMaximum(num)
        self.close_button = QPushButton('Close')
        self.close_button.clicked.connect(self.close)
        self.label = QLabel(self.progress_label)
        layout.addWidget(self.label)
        layout.addWidget(self.progress)
        layout.addWidget(self.close_button)
        self.close_button.hide()

        if not num:
            self.progress.setMaximum(1)
            self.progress.setValue(1)
            self.finished()

    @pyqtSlot(int, name='setProgress')
    def set_progress(self, value: int):
        self.progress.setValue(value)

    @pyqtSlot(name='finished')
    def finished(self):
        self.label.setText(self.finish_label)
        self.progress.setValue(self.progress.maximum())
        self.close_button.show()



import pickle
from dataclasses import dataclass
from collections import defaultdict
import logging
from pathlib import Path

from .keys import (AbstractKey, FolderKey, FolderH5Key, FolderPathKey,
                   ImageKey, ImagePathKey, H5_FORMAT, RemoveWeakrefs,
                   AVAILABLE_IMAGE_FORMATS)


class ProjectRootKey(FolderKey):
    def __init__(self, *, project_path: Path = None, h5source_path: Path = None):
        super().__init__(parent=None)
        self.path: Path = project_path
        self.h5_path: Path = h5source_path

        self._valid = True

        if self.h5_path:
            self._project_name = self.h5_path.name
        elif self.path:
            self._project_name = self.path.name
        else:
            self._project_name = 'EMPTY PROJECT'
            self._valid = False

        self.update()

    def is_valid(self) -> bool:
        return self._valid

    def __eq__(self, other):
        if isinstance(other, ProjectRootKey):
            return self.h5_path == other.h5_path and self.path == other.path

    def _file_key(self) -> str:
        return self.name

    @property
    def name(self):
        return self._project_name

    @property
    def parent(self):
        return

    def add_path(self, path: Path):
        if path.is_dir():
            key = FolderPathKey(self, path=path)
            self._folder_children.append(key)
            return key
        elif path.is_file() and path.suffix in AVAILABLE_IMAGE_FORMATS:
            key = ImagePathKey(self, path=path)
            self._image_children.append(key)
            return key
        elif path.is_file() and path.suffix in H5_FORMAT:
            key = FolderH5Key(self, h5path=path)
            self._folder_children.append(key)
            return key

    def remove_key(self, key: AbstractKey):
        if isinstance(key, FolderKey):
            try:
                self._folder_children.remove(key)
            except ValueError:
                return
        elif isinstance(key, ImageKey):
            try:
                self._image_children.remove(key)
            except ValueError:
                return


@dataclass
class H5ProjectConfig:
    save_to_h5: bool = False


class ProjectStructure(object):
    log = logging.getLogger(__name__)

    def __init__(self):
        self._root: ProjectRootKey = None
        self.config = defaultdict(lambda: False)
        self.path: Path = None

    @property
    def project_opened(self):
        return self._root is not None

    @property
    def root(self) -> ProjectRootKey:
        return self._root

    def save(self, restore: bool = True):
        if self.path and self.root and self.path.is_dir():
            with RemoveWeakrefs(self.root, restore=restore):
                with open(str(self.path.resolve() / 'project_structure'), 'wb') as f:
                    pickle.dump(self.root, f)

    def open_project(self, path: Path):
        self.close_project()

        self.path = path

        if self.path.is_file():
            raise ValueError(f'Project path is a file: {self.path.resolve()}')

        self.path.mkdir(exist_ok=True, parents=True)

        pickle_file = self.path / 'project_structure'

        if pickle_file.is_file():
            try:
                with open(str(pickle_file.resolve()), 'rb') as f:
                    self._root = pickle.load(f)
                RemoveWeakrefs.restore(self._root)
            except Exception as err:
                pickle_file.unlink()
                self.log.exception(err)

        if not self._root:
            self._root = ProjectRootKey(project_path=self.path)

    def save_and_close(self):
        self.save(False)
        self.close_project()

    def close_project(self):
        self._root = None
        self.path = None
        self.config = defaultdict(lambda: False)



from PyQt5.QtWidgets import QWidget, QGridLayout, QLabel, QFrame
from PyQt5.QtCore import Qt, QObject, QPoint, pyqtSignal, pyqtSlot
from .spinner import Spinner
from .labels import Label, LabelButton

NOTIFICATION_BACKGROUND_COLOR = 'rgb(80, 80, 80)'


class PopUpWrapper(QObject):
    WIDTH = 400
    HEIGHT = 90
    PADDING = 10

    def __init__(self, widget: QWidget):
        super().__init__(widget)
        self.parent = widget
        self.__popups = []
        self.__popups_dict = {}
        self.__monkeypatch_parent_methods(widget)

    @pyqtSlot(name='movePopup')
    def _move_popups(self):
        win_width, win_height = self.parent.size().width(), self.parent.size().height()
        win_point = self.parent.mapToGlobal(QPoint(win_width, win_height))
        for i, popup in enumerate(self.__popups):
            popup_w, popup_h = self.WIDTH + self.PADDING, (self.HEIGHT + self.PADDING) * (i + 1)
            p = win_point - QPoint(popup_w, popup_h)
            popup.move(p)
            if win_width - popup_w < 0 or win_height - popup_h < 0:
                popup.hide()
            else:
                popup.show()

    def add_notification(self, context: str or QWidget, name=None):
        if not context:
            return
        notification = Notification(self.parent, context)
        notification.sigCloseClicked.connect(self.remove_notification)

        if name is not None:
            self.__popups_dict[name] = notification

        self.__popups.append(notification)
        self._move_popups()

    @pyqtSlot(QWidget, name='removeWidget')
    def remove_notification(self, widget: QWidget):
        try:
            self.__popups.remove(widget)
            widget.deleteLater()
            self._move_popups()
        except ValueError:
            pass
        self.__popups_dict = {k: v for k, v in self.__popups_dict.items() if v is not widget}

    def remove_by_name(self, name):
        if name in self.__popups_dict:
            self.remove_notification(self.__popups_dict.pop(name))

    def clear(self):
        for widget in self.__popups:
            try:
                widget.deleteLater()
            except RuntimeError:
                pass
        self.__popups = []
        self.__popups_dict = {}

    def __monkeypatch_parent_methods(self, parent: QWidget):
        old_resize = parent.resizeEvent
        old_move = parent.moveEvent

        def new_resize(event):
            self._move_popups()
            old_resize(event)

        def new_move(event):
            self._move_popups()
            old_move(event)

        parent.resizeEvent = new_resize
        parent.moveEvent = new_move


class CloseButton(LabelButton):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        font = self.font()
        font.setBold(True)
        font.setPointSize(10)
        self.setFont(font)

    def set_visible(self, enable: bool = True):
        if enable:
            self.setStyleSheet('QLabel {color: #808080}')
        else:
            self.setStyleSheet('QLabel {color: %s}' % NOTIFICATION_BACKGROUND_COLOR)


class Notification(QFrame):
    sigCloseClicked = pyqtSignal(QWidget)

    def __init__(self, parent, context: str or QWidget):
        super().__init__(parent, Qt.Window | Qt.FramelessWindowHint)
        self._init_ui(context)
        self.show()

    def enterEvent(self, a0) -> None:
        self.button.set_visible(True)
        super().enterEvent(a0)

    def leaveEvent(self, a0) -> None:
        self.button.set_visible(False)
        super().leaveEvent(a0)

    def _init_ui(self, context: str or QWidget):
        if isinstance(context, str):
            context = QLabel(context, self)
        else:
            context.setParent(self)

        self.setFixedSize(PopUpWrapper.WIDTH, PopUpWrapper.HEIGHT)
        self.setStyleSheet(f"background-color: {NOTIFICATION_BACKGROUND_COLOR}")
        self.setContentsMargins(5, 5, 5, 0)

        layout = QGridLayout(self)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(0)
        layout.addWidget(self._get_close_button(), 0, 1, alignment=Qt.AlignRight | Qt.AlignTop)
        layout.addWidget(context, 0, 0, 3, 1)

    def _get_close_button(self):
        close_button = CloseButton('x', self)
        close_button.sigClicked.connect(lambda: self.sigCloseClicked.emit(self))
        close_button.set_visible(False)
        self.button = close_button
        return close_button


class TextNotification(QWidget):
    def __init__(self, title: str, text: str, spinner: bool = False, parent=None):
        super().__init__(parent)
        layout = QGridLayout(self)
        layout.setSpacing(5)

        if spinner:
            layout.addWidget(Spinner(self, 20), 0, 0)
            layout.addWidget(Label(title, self, 8.5, True), 0, 1)
            layout.addWidget(Label(text, self, 8.5), 1, 0, 1, 2)
        else:
            layout.addWidget(Label(title, self, 8.5, True), 0, 0)
            layout.addWidget(Label(text, self, 8.5), 1, 0)



from h5py import Group

from .npy_file_manager import _ReadNpy


class _ReadPolarImage(_ReadNpy):
    NAME = 'polar_images'

    @staticmethod
    def _get_h5(h5group: Group, key):
        if 'polar_image' in h5group.keys():
            return h5group['polar_image'][()]

    @staticmethod
    def _set_h5(h5group: Group, key, image):
        _ReadPolarImage._del_h5(h5group)
        h5group.create_dataset('polar_image', data=image)

    @staticmethod
    def _del_h5(h5group: Group, key):
        if 'image' in h5group.keys():
            del h5group['polar_image']



from pathlib import Path

import numpy as np

from .object_file_manager import _ObjectFileManager


class _ReadNpy(_ObjectFileManager):
    @staticmethod
    def _set_pickle(path: Path, value):
        np.save(str(path.resolve()) + '.npy', value)

    @staticmethod
    def _get_pickle(path: Path):
        if path.is_file():
            return np.load(str(path.resolve()) + '.npy', allow_pickle=True)




# -*- coding: utf-8 -*-
import logging

from PyQt5.QtWidgets import (QAbstractButton, QGraphicsDropShadowEffect,
                             QHBoxLayout, QLabel, QWidget)
from PyQt5.QtGui import (QIcon, QPainter, QBrush, QColor)
from PyQt5.QtCore import Qt, QRect, pyqtSignal

from ..tools import Icon

logger = logging.getLogger(__name__)


class RoundedPushButton(QAbstractButton):
    def __init__(self,
                 parent=None, *,
                 text: str = '',
                 icon: QIcon = None,
                 radius: int = 20,
                 background_color: QColor or str = 'white'):
        self._radius = radius
        self._text = text
        self._icon = icon
        if isinstance(background_color, str):
            if background_color == 'transparent':
                background_color = QColor(0, 0, 0, 0)
            else:
                background_color = QColor(background_color)
        self._background_color = background_color
        super().__init__(parent)
        self.__init_shadow__()
        self._pressed = False
        if self._icon:
            self.setFixedSize(self._radius, self._radius)

    def __init_shadow__(self):
        self.shadow = QGraphicsDropShadowEffect()
        self.shadow.setBlurRadius(self._radius)
        self.shadow.setXOffset(3)
        self.shadow.setYOffset(3)
        self.setGraphicsEffect(self.shadow)

    def _get_painter(self):
        painter = QPainter(self)
        brush = QBrush()
        brush.setColor(self._background_color)
        brush.setStyle(Qt.SolidPattern)
        painter.setBrush(brush)
        painter.setPen(Qt.NoPen)
        painter.setRenderHint(QPainter.Antialiasing)
        return painter

    def paintEvent(self, QPaintEvent):
        painter = self._get_painter()
        # TODO: fix icon position and size bug when pressed
        if self._pressed:
            rect = QRect(0, 3, painter.device().width(),
                         int(painter.device().height() * 0.9))
        else:
            rect = QRect(0, 0, painter.device().width(),
                         painter.device().height())
        painter.drawRoundedRect(rect, self._radius, self._radius)
        if self._icon:
            pixmap = self._icon.pixmap(self.size())
            painter.drawPixmap(0, 0, pixmap)

    def mousePressEvent(self, QMouseEvent):
        super().mousePressEvent(QMouseEvent)
        if QMouseEvent.button() == Qt.LeftButton:
            self._pressed = True
            self.shadow.setEnabled(False)

    def mouseReleaseEvent(self, QMouseEvent):
        self._pressed = False
        super().mouseReleaseEvent(QMouseEvent)
        self.shadow.setEnabled(True)
        self.shadow.setBlurRadius(self._radius)


class ConfirmButton(QWidget):
    clicked = pyqtSignal()

    def __init__(self, icon: QIcon, parent=None, radius: float = 30,
                 smaller_radius: float = 25,
                 text: str = ''):
        QWidget.__init__(self, parent)
        self._icon = icon
        self._text = text
        self._status = None
        self.__init_ui__(radius, smaller_radius)
        self.hide_and_show('icon')

    def __init_ui__(self, radius, smaller_radius):
        layout = QHBoxLayout(self)
        self.setLayout(layout)

        self.icon_widget = RoundedPushButton(self, icon=self._icon,
                                             radius=radius)
        self.icon_widget.clicked.connect(self.on_clicked)
        layout.addWidget(self.icon_widget)

        self.label_widget = QLabel(self._text)
        self.confirm = RoundedPushButton(
            icon=Icon('confirm'), radius=smaller_radius)
        self.confirm.clicked.connect(self.on_clicked)
        self.confirm.clicked.connect(self.clicked.emit)
        self.decline = RoundedPushButton(
            icon=Icon('delete'), radius=smaller_radius)
        self.decline.clicked.connect(self.on_clicked)
        layout.addWidget(self.label_widget)
        layout.addWidget(self.confirm)
        layout.addWidget(self.decline)

    def on_clicked(self):
        self.hide_and_show()

    def hide_and_show(self, new_status: str = None):
        if new_status in 'icon question'.split():
            self._status = new_status
        else:
            self._status = 'icon' if self._status == 'question' else 'question'

        if self._status == 'icon':
            self.label_widget.hide()
            self.confirm.hide()
            self.decline.hide()
            self.icon_widget.show()
        else:
            self.label_widget.show()
            self.confirm.show()
            self.decline.show()
            self.icon_widget.hide()


class DeleteButton(ConfirmButton):
    def __init__(self, parent=None, radius: float = 30,
                 smaller_radius: float = 25, text: str = 'Delete?'):
        ConfirmButton.__init__(self, Icon('delete'), parent,
                               radius, smaller_radius, text=text)


class InfoButton(RoundedPushButton):
    def __init__(self, info, parent=None):
        RoundedPushButton.__init__(self, parent, icon=Icon('info'))
        self.setToolTip(info)



from enum import Enum
from dataclasses import dataclass


class RangeStrategyType(Enum):
    fixed = 'Fixed'
    adjust = 'Adjust'


@dataclass
class RangeStrategy:
    range_factor: float = 0
    strategy_type: RangeStrategyType = RangeStrategyType.adjust
    is_default: bool = True



from datetime import datetime as dt
from enum import Enum
import sys
from typing import Dict
import weakref
import gc
import logging

from PyQt5.QtCore import QObject, pyqtSignal, pyqtSlot

from .utils import SingletonMeta

logger = logging.getLogger(__name__)


class ObjectStatus(Enum):
    EXISTS = 0
    SAFELY_DELETED = 1
    DELETED_BY_PYTHON = 2
    DELETED_BY_C = 3


class TrackSignals(QObject):
    sigObjectAdded = pyqtSignal(str)
    sigStatusUpdated = pyqtSignal(list)


class ObjectTracker(QObject):
    def __init__(self, obj: QObject, name: str):
        super().__init__()
        self.name: str = f'{obj.__class__.__name__}: {name}'
        self.id: str = f'{name}: {id(obj)}, {dt.now().timestamp()}'
        self.obj_ref = weakref.ref(obj)
        self.status: ObjectStatus = ObjectStatus.EXISTS
        self._deleted_by_c: bool = False
        self._deleted_by_python: bool = False
        obj.destroyed.connect(self._on_destroyed)
        self.update_status()

    def update_status(self) -> ObjectStatus:
        try:
            obj = self.obj_ref()
            if obj is None:
                self._deleted_by_python = True
        except LookupError:
            logger.debug('LookupError in tracker')
            self._deleted_by_python = True
        except RuntimeError:
            logger.debug('RuntimeError in tracker')
            self._deleted_by_c = True

        if self._deleted_by_c and self._deleted_by_python:
            self.status = ObjectStatus.SAFELY_DELETED
        elif self._deleted_by_c and not self._deleted_by_python:
            self.status = ObjectStatus.DELETED_BY_C
        elif not self._deleted_by_c and self._deleted_by_python:
            self.status = ObjectStatus.DELETED_BY_PYTHON
        else:
            self.status = ObjectStatus.EXISTS

        return self.status

    def get_referrers(self):
        obj = self.get_obj()
        if obj:
            logger.debug(f'Number of refs = {sys.getrefcount(obj)}')
            return gc.get_referrers([obj])
        else:
            logger.debug('Object is None!')

    def get_obj(self):
        try:
            return self.obj_ref()
        except Exception as err:
            logger.exception(err)

    def is_tracked(self) -> bool:
        obj = self.get_obj()
        if obj:
            return gc.is_tracked(obj)
        else:
            return False

    @pyqtSlot(name='onDestroyed')
    def _on_destroyed(self, *args):
        logger.debug('destroyed signal in tracker')
        self._deleted_by_c = True


class TrackQObjects(metaclass=SingletonMeta):
    ObjectStatus: ObjectStatus = ObjectStatus

    def __init__(self):
        self.signals: TrackSignals = TrackSignals()
        self.objects: Dict[str, ObjectTracker] = {}

    def add_object(self, obj: QObject, name: str = ''):
        obj_tracker = ObjectTracker(obj, name)
        self.objects[obj_tracker.id] = obj_tracker
        self.signals.sigObjectAdded.emit(obj_tracker.id)

    def update(self, key: str = None, *, emit: bool = True):
        if key:
            self._update(key)
        else:
            updated_keys = []
            for key, obj in self.objects.items():
                if obj.status != obj.update_status():
                    updated_keys.append(key)
            if emit and updated_keys:
                self.signals.sigStatusUpdated.emit(updated_keys)

    def _update(self, key: str):
        try:
            if self.objects[key].status != self.objects[key].update_status():
                self.signals.sigStatusUpdated.emit([key])
        except KeyError:
            pass

    def clear_deleted(self, status: ObjectStatus = ObjectStatus.SAFELY_DELETED):
        to_delete = []
        for key, obj in self.objects.items():
            if obj.status == status:
                to_delete.append(key)

        for key in to_delete:
            obj = self.objects.pop(key)
            obj.deleteLater()
        return to_delete



from enum import Enum
from abc import abstractmethod
from typing import Dict

import numpy as np

from .utils import Roi

__all__ = ['BackgroundType', 'BACKGROUNDS', 'Background', 'LinearBackground', 'ConstantBackground']


class BackgroundType(Enum):
    constant = 'Constant'
    linear = 'Linear'


class Background(object):
    PARAM_NAMES: tuple = ()
    NUM: int = 0
    TYPE: BackgroundType = None

    is_default: bool = True

    AMP_PADDING: float = 1.1

    @staticmethod
    @abstractmethod
    def _bounds(x: np.ndarray, y: np.ndarray) -> tuple:
        pass

    def bounds(self, x: np.ndarray, y: np.ndarray, roi: Roi = None):
        return self._bounds(x, y)

    @abstractmethod
    def __call__(self, x: np.ndarray, *params) -> np.ndarray:
        pass

    @abstractmethod
    def amp_bounds(self, x: np.ndarray, y: np.ndarray, params: list) -> tuple:
        pass


class ConstantBackground(Background):
    PARAM_NAMES = ('constant background',)
    NUM = 1
    TYPE = BackgroundType.constant

    @staticmethod
    def _bounds(x: np.ndarray, y: np.ndarray) -> tuple:
        min_y = y.min()

        return [min_y], [y.max()], [min(0, min_y)]

    def __call__(self, x: np.ndarray, *params) -> np.ndarray:
        return params[-1] * np.ones_like(x)

    def amp_bounds(self, x: np.ndarray, y: np.ndarray, params: list) -> tuple:
        amp = y.max() * self.AMP_PADDING - params[0]
        return amp, max(y.max(), amp * 2), min(amp, 0)


class LinearBackground(Background):
    PARAM_NAMES = ('background (slope)', 'background (level)')
    NUM = 2
    TYPE = BackgroundType.linear

    @staticmethod
    def _bounds(x: np.ndarray, y: np.ndarray) -> tuple:
        dx = (x[-1] - x[0]) or 1
        b1 = (y[-1] - y[0]) / dx
        b2 = y[0] - x[0] * b1

        return [b1, b2], [b1 + abs(b1) / 2, b2 + abs(b2) / 2], [b1 - abs(b1) / 2, b2 - abs(b2) / 2]

    def __call__(self, x: np.ndarray, *params) -> np.ndarray:
        return params[-2] * x + params[-1]

    def amp_bounds(self, x: np.ndarray, y: np.ndarray, params: list) -> tuple:
        amp = y.max() * self.AMP_PADDING - (y[0] + y[-1]) / 2
        return amp, max(y.max(), amp * 2), min(amp, 0)


BACKGROUNDS: Dict[BackgroundType, Background.__class__] = {
    BackgroundType.constant: ConstantBackground,
    BackgroundType.linear: LinearBackground,
}



from PyQt5.QtWidgets import QToolBar


class ToolBar(QToolBar):
    def __init__(self, name: str, parent=None, color: str = None, *,
                 disable_hide: bool = True, movable: bool = False):
        super().__init__(name, parent)
        if color:
            self.setStyleSheet(f'background-color: {color};')
        if disable_hide:
            self.toggleViewAction().setEnabled(False)
        self.setMovable(movable)


class BlackToolBar(ToolBar):
    def __init__(self, name: str, parent=None, **kwargs):
        super().__init__(name, parent, 'black', **kwargs)



from abc import abstractmethod
from ..app.rois.roi import Roi


class AbstractRoi:
    def __init__(self, roi: Roi):
        self.roi = roi

    @abstractmethod
    def set_color(self, color):
        pass

    @abstractmethod
    def fix(self):
        pass

    @abstractmethod
    def unfix(self):
        pass

    @abstractmethod
    def show_roi(self):
        pass

    @abstractmethod
    def hide_roi(self):
        pass

    



# -*- coding: utf-8 -*-

import logging

from scipy.signal import find_peaks

from PyQt5.QtGui import QColor
from PyQt5.QtCore import QPointF

from ..basic_widgets import (ConfirmButton, DrawRoiController,
                             RoundedPushButton, PlotBC,
                             BlackToolBar, BasicInputParametersWidget)
from ...app.app import App
from ..tools import Icon
from ..roi_widgets.abstract_roi_holder import AbstractRoiHolder
from ..roi_widgets.roi_1d_widget import Roi1D
from ..tools import show_error

logger = logging.getLogger(__name__)


class RadialProfileWidget(AbstractRoiHolder, PlotBC):

    def __init__(self, parent=None):
        AbstractRoiHolder.__init__(self, 'RadialProfile')
        PlotBC.__init__(self, profile=App().radial_profile, parent=parent)
        self.app = App()
        self.app.geometry_holder.sigScaleChanged.connect(self._update_axis)
        self.app.image_holder.sigEmptyImage.connect(self.clear_plot)
        self._init_radial_toolbars()
        self._fit_params_dict: dict = PeaksSetupWindow.get_config()
        self._peaks_setup = None
        self._draw_roi = RadialDrawRoi(self.view_box, self)
        self._draw_roi.sigCreateRoi.connect(self.app.roi_dict.add_roi)
        self._draw_roi.sigMoveRoi.connect(self.app.roi_dict.move_roi)

        # self.image_view.plot_item.setTitle('Radial Profile')
        self.image_view.plot_item.getAxis('bottom').setLabel('|Q|', color='white', font_size='large')
        self.image_view.plot_item.getAxis('left').setLabel('Intensity', color='white', font_size='large')

        self.sigBackgroundChanged.connect(self.profile.save_state)

    def _init_radial_toolbars(self):

        fit_toolbar = BlackToolBar('Fitting', self)
        self.addToolBar(fit_toolbar)

        find_peaks_widget = ConfirmButton(Icon('find'), text='Find peaks?')
        find_peaks_widget.label_widget.setStyleSheet(
            'QLabel { color : white ; }')
        find_peaks_widget.clicked.connect(self.find_peaks)
        fit_toolbar.addWidget(find_peaks_widget)
        #
        # fit_peaks_widget = ConfirmButton(Icon('fit'), text='Fit selected peaks?')
        # fit_peaks_widget.label_widget.setStyleSheet(
        #     'QLabel { color : white ; }')
        # fit_peaks_widget.clicked.connect(self.fit_selected)
        # fit_toolbar.addWidget(fit_peaks_widget)
        #
        setup_action = fit_toolbar.addAction(Icon('setup'), 'Fit setup')
        setup_action.triggered.connect(self.open_peaks_setup)

        segments_toolbar = BlackToolBar('Segments', self)
        self.addToolBar(segments_toolbar)

        create_roi_widget = RoundedPushButton(
            icon=Icon('add'), radius=30)
        create_roi_widget.clicked.connect(lambda: self.app.roi_dict.create_roi())
        segments_toolbar.addWidget(create_roi_widget)

        delete_selected_widget = ConfirmButton(
            Icon('delete'), text='Delete selected roi?')
        delete_selected_widget.clicked.connect(self.app.roi_dict.delete_selected_roi)
        delete_selected_widget.label_widget.setStyleSheet(
            'QLabel { color : white ; }')

        segments_toolbar.addWidget(delete_selected_widget)
        fix_all = RoundedPushButton(icon=Icon('fix_all'), radius=120, background_color=QColor(0, 0, 0, 0))
        fix_all.setFixedWidth(60)
        fix_all.setFixedHeight(30)
        fix_all.clicked.connect(self.app.roi_dict.fix_all)
        segments_toolbar.addWidget(fix_all)
        unfix_all = RoundedPushButton(icon=Icon('unfix_all'), radius=120, background_color=QColor(0, 0, 0, 0))
        unfix_all.setFixedWidth(60)
        unfix_all.setFixedHeight(30)
        unfix_all.clicked.connect(self.app.roi_dict.unfix_all)
        segments_toolbar.addWidget(unfix_all)

    # def _on_scale_changed(self):
    #     self.update_x_axis()
    #
    #
    def _update_axis(self):
        self.x = App().geometry.r_axis

    def find_peaks(self):
        if self.y is None or self.x is None:
            return
        if self._fit_params_dict.get('sigma_find', None) is not None:
            self.set_sigma(self._fit_params_dict['sigma_find'])
        peaks = find_peaks(self.y)[0]
        max_num = self._fit_params_dict.get('max_peaks_number', 40)
        if len(peaks) > max_num:
            show_error(
                'Maximum number of peaks exceeded.',
                error_title='Error',
                info_text=f'Number of found peaks ({len(peaks)}) exceeds the maximum number {max_num}. '
                          f'Consider increasing sigma for smoothing the radial profile.')
            return
        w = self._fit_params_dict.get('init_width', 30) * App().geometry.scale
        for r in peaks:
            self._roi_dict.create_roi(radius=self.x[r], width=w)

    def _make_roi_widget(self, roi):
        roi_widget = Roi1D(roi)
        self.image_view.plot_item.addItem(roi_widget)
        return roi_widget

    def _delete_roi_widget(self, roi_widget):
        self.image_view.plot_item.removeItem(roi_widget)

    def open_peaks_setup(self):
        self._peaks_setup = PeaksSetupWindow()
        self._peaks_setup.apply_signal.connect(self.set_fit_parameters)
        self._peaks_setup.close_signal.connect(self.close_peaks_setup)
        self._peaks_setup.show()

    def set_fit_parameters(self, params: dict):
        self._fit_params_dict = params

    def close_peaks_setup(self):
        self._peaks_setup = None


class RadialDrawRoi(DrawRoiController):
    def _update_roi(self, point: QPointF):
        r1, r2 = self._init_point.x(), point.x()
        self._roi.radius, self._roi.width = (r1 + r2) / 2, abs(r2 - r1)


class PeaksSetupWindow(BasicInputParametersWidget):
    P = BasicInputParametersWidget.InputParameters

    PARAMETER_TYPES = (P('max_peaks_number',
                         'Maximum number of peaks',                         int, 'Do not recommended to put high numbers'),
                       P('init_width', 'Peaks width', float,
                         'Gaussian fitting will start with this number'),
                       P('sigma_find', 'Sigma to find peaks', float,
                         'Default sigma value for gaussian smooth\n'
                         'applied before initial peaks finding to\n'
                         'avoid noise peaks. To use current lambda, \n'
                         'leave empty.', True),
                       P('sigma_fit', 'Sigma to find peaks', float,
                         'Default sigma value for gaussian smooth\n'
                         'applied before gaussian fitting of \n'
                         'found peaks. To use current lambda, \n'
                         'leave empty.', True)
                       )

    NAME = 'Fitting parameters'

    DEFAULT_DICT = dict(max_peaks_number=20, init_width=30, sigma_find=8)



# -*- coding: utf-8 -*-
import logging
from copy import deepcopy
from abc import abstractmethod
from typing import NamedTuple

from PyQt5.QtWidgets import (QWidget, QLineEdit, QHBoxLayout,
                             QLabel, QFormLayout,
                             QPushButton, QRadioButton)
from PyQt5.QtGui import QIntValidator
from PyQt5.QtCore import Qt, pyqtSignal

from ...app.app import App
from .buttons import InfoButton
from ..tools import Icon, center_widget, validate_scientific_value

logger = logging.getLogger(__name__)


class AbstractInputParametersWidget(QWidget):
    class InputParameters(NamedTuple):
        name: str
        label: str
        type: type
        info: str = None
        none: str = False

    @property
    @abstractmethod
    def PARAMETER_TYPES(self):
        pass

    @property
    @abstractmethod
    def NAME(self):
        pass

    @property
    @abstractmethod
    def DEFAULT_DICT(self):
        return {}

    @classmethod
    def get_config(cls):
        d = deepcopy(cls.DEFAULT_DICT)
        config_d = App().fm.config[cls.NAME] or {}
        d.update(config_d)
        return d

    def __init__(self, parent=None, params_dict: dict = None):
        super().__init__(parent)
        self.setWindowIcon(Icon('setup'))
        self.setWindowTitle(self.NAME)
        self.default_dict = params_dict or self.get_config()

    def get_parameters_dict(self):
        parameters_dict = dict()
        for p in self.PARAMETER_TYPES:
            value = getattr(self, p.name)
            if not p.none and value is None:
                return
            parameters_dict[p.name] = value
        return parameters_dict

    def save_to_config(self, parameters_dict: dict):
        App().fm.config[self.NAME] = parameters_dict

    def _get_layout(self,
                    input_parameter: 'InputParameters'):
        current_value = self.default_dict.get(input_parameter.name, '')
        if current_value is None:
            current_value = ''
        label_widget = QLabel(input_parameter.label)
        input_widget = QLineEdit(str(current_value))
        if input_parameter.type is int:
            input_widget.setValidator(QIntValidator())
        layout = QHBoxLayout()
        layout.addWidget(label_widget, Qt.AlignHCenter)
        layout.addWidget(input_widget, Qt.AlignHCenter)
        if input_parameter.info:
            info_button = InfoButton(input_parameter.info)
            layout.addWidget(info_button, Qt.AlignLeft)

        def get_input(s):
            return validate_scientific_value(input_widget, input_parameter.type, input_parameter.none)

        setattr(AbstractInputParametersWidget, input_parameter.name,
                property(get_input))
        return layout


class BasicInputParametersWidget(AbstractInputParametersWidget):
    apply_signal = pyqtSignal(dict)
    close_signal = pyqtSignal()

    def __init__(self, parent=None, params_dict: dict = None):
        super().__init__(parent, params_dict)
        self.form = QFormLayout()
        for p in self.PARAMETER_TYPES:
            self.form.addRow(self._get_layout(p))
        self.save_button = self._init_save_button()
        self.apply_button = self._init_apply_button()
        self.cancel_button = self._init_cancel_button()
        self.form.addWidget(self.save_button)
        self.form.addWidget(self.apply_button)
        self.form.addWidget(self.cancel_button)
        self.setLayout(self.form)
        center_widget(self)

    def _init_apply_button(self):
        apply_button = QPushButton('Apply')

        apply_button.clicked.connect(self.on_apply)
        return apply_button

    def _init_cancel_button(self):
        cancel_button = QPushButton('Cancel')
        cancel_button.clicked.connect(self.close)
        return cancel_button

    def on_apply(self, *args):
        parameters_dict = self.get_parameters_dict()
        if parameters_dict is not None:
            self.apply_signal.emit(parameters_dict)
            logger.debug(parameters_dict)
            if self.save_button.isChecked():
                self.save_to_config(parameters_dict)
            self.close()

    def close(self):
        self.close_signal.emit()
        AbstractInputParametersWidget.close(self)

    @staticmethod
    def _init_save_button():
        save_button = QRadioButton('Save parameters')
        save_button.setChecked(True)
        return save_button



# -*- coding: utf-8 -*-
import platform
import subprocess
import os


def launch_detached():
    try:
        if platform.system() == 'Windows':
            subprocess.Popen('giwaxs_gui', creationflags=subprocess.DETACHED_PROCESS)
        else:
            subprocess.Popen(['nohup', 'giwaxs_gui'], shell=False, stdout=None, stderr=None, preexec_fn=os.setpgrp)

    except subprocess.CalledProcessError:
        return -1



# -*- coding: utf-8 -*-
import logging
from enum import Enum

import numpy as np

from PyQt5.QtWidgets import (QMainWindow, QWidget,
                             QFrame, QHBoxLayout,
                             QVBoxLayout, QPushButton)
from PyQt5.QtGui import QColor, QPen
from PyQt5.QtCore import Qt, pyqtSignal

from pyqtgraph import GraphicsLayoutWidget, LinearRegionItem
from .sliders import LabeledSlider
from .toolbars import BlackToolBar
from ..basic_widgets import RoundedPushButton
from ..tools import Icon, show_error
from ...app.profiles import BasicProfile

from ..tools import get_pen, center_widget

logger = logging.getLogger(__name__)


class Custom1DPlot(GraphicsLayoutWidget):
    @property
    def view_box(self):
        return self.plot_item.vb

    def __init__(self, *args, parent=None):
        super().__init__(parent)
        self.plot_item = self.addPlot()
        self.plot_item.setMenuEnabled(False)
        self.plot = self.plot_item.plot(*args)
        self.set_pen()

    def set_data(self, *args):
        self.plot.setData(*args)

    def clear_plot(self):
        self.plot.clear()

    def set_x(self, x):
        x = np.array(x)
        y = self.plot.yData
        if y is None or y.shape != x.shape:
            return
        self.plot.setData(x, y)

    def set_pen(self, width: int = 3, color: str or QColor = 'white'):
        if isinstance(color, str):
            color = QColor(color)
        pen = QPen(color)
        pen.setStyle(Qt.SolidLine)
        pen.setWidth(width)
        pen.setCapStyle(Qt.RoundCap)
        pen.setJoinStyle(Qt.RoundJoin)
        pen.setCosmetic(True)
        self.plot.setPen(pen)


class Smooth1DPlot(QMainWindow):
    _MaximumSliderWidth = 200
    _MaximumSliderHeight = 30

    sigSigmaChanged = pyqtSignal(float)

    @property
    def view_box(self):
        return self.image_view.view_box

    def __init__(self, profile: BasicProfile, parent=None):
        super().__init__(parent)
        self.profile = profile
        self.image_view = Custom1DPlot()
        self.setCentralWidget(self.image_view)
        self._init_toolbars()

    @property
    def x(self):
        return self.profile.x

    @x.setter
    def x(self, value):
        self.profile.x = value
        self.plot()

    @property
    def y(self) -> np.ndarray or None:
        return self.profile.y

    def _init_toolbars(self):
        param_toolbar = BlackToolBar('Parameters', self)
        self.addToolBar(param_toolbar)

        self._init_sigma_slider(param_toolbar)

    def _init_sigma_slider(self, toolbar):
        sigma_slider = LabeledSlider('𝞼', (0, 10), self.profile.sigma,
                                     decimals=2)

        sigma_slider.valueChanged.connect(self.update_sigma)

        self.sigma_slider = sigma_slider
        frame = QFrame()
        layout = QHBoxLayout()
        frame.setLayout(layout)
        frame.setGeometry(0, 0, self._MaximumSliderWidth, toolbar.height() * 0.9)
        layout.addWidget(sigma_slider, alignment=Qt.AlignLeft)
        toolbar.addWidget(frame)

    def set_sigma(self, value: float):
        self.update_sigma(value)
        self.sigma_slider.set_value(value, change_bounds=True)

    def update_sigma(self, value: float):
        self.profile.set_sigma(value)
        self.plot()
        self.sigSigmaChanged.emit(self.profile.sigma)

    def auto_range(self):
        self.image_view.plot_item.autoRange()

    def plot(self):
        if self.x is not None and self.y is not None:
            self.image_view.set_data(self.x, self.y)
            self.auto_range()

    def clear_plot(self):
        self.image_view.clear_plot()


class PlotBC(Smooth1DPlot):
    sigBackgroundChanged = pyqtSignal()

    def __init__(self, profile: BasicProfile, parent=None):
        super().__init__(profile, parent)
        self._status = BaseLineStatus.no_baseline
        self.baseline_plot = self.image_view.plot_item.plot()
        self._init_roi()
        self._baseline_setup_widget = BaseLineSetup(self, self._status, **self.profile.get_parameters())
        self.profile.sigDataUpdated.connect(self.update_plot)

    @property
    def y(self):
        if self._status == BaseLineStatus.baseline_subtracted and self.profile.baseline is not None:
            return self.profile.y - self.profile.baseline
        else:
            return self.profile.y

    def update_data(self, *args, **kwargs):
        self.profile.update_data(*args, **kwargs)

    def is_shown(self, shown: bool):
        self.profile.is_shown = shown

    def _init_toolbars(self):
        super()._init_toolbars()

        baseline_toolbar = BlackToolBar('Baseline Correction')
        self.addToolBar(baseline_toolbar)

        baseline_button = RoundedPushButton(parent=baseline_toolbar, icon=Icon('baseline'),
                                            radius=30)
        baseline_button.clicked.connect(self.open_baseline_setup)
        baseline_toolbar.addWidget(baseline_button)

    def _init_roi(self):
        self._roi = LinearRegionItem()
        self._roi.hide()
        self._roi.setBrush(QColor(255, 255, 255, 50))
        self.image_view.plot_item.addItem(self._roi)

    def open_baseline_setup(self):
        if self.y is None:
            return
        setup = self._baseline_setup_widget

        if self.profile.x_range is None:
            self._set_default_bounds()
        self._roi.setRegion(self.profile.x_range)

        if self.profile.baseline is None:
            self._set_status(BaseLineStatus.no_baseline)
        # elif self._status == BaseLineStatus.no_baseline:
        #     self._set_status(BaseLineStatus.baseline_subtracted)
        # else:
        #     self._set_status(self._status)

        # self.plot()

        setup.set_parameters(self.profile.get_parameters())

        setup.calculate_signal.connect(self._on_calculate_baseline)
        setup.subtract_signal.connect(self._on_subtracting_baseline)
        setup.restore_signal.connect(self._on_restoring_data)
        setup.close_signal.connect(self._on_closing_setup)
        setup.show()
        self._roi.show()

    # def show_baseline(self):
    #     if (self.profile.baseline is None or self._status == BaseLineStatus.baseline_calculated or
    #             self._status == BaseLineStatus.baseline_restored):
    #         return
    #     self._on_restoring_data()

    def update_plot(self):
        self.sigma_slider.set_value(self.profile.sigma, True)

        if self.profile.baseline is None:
            self.clear_baseline()
        else:
            self._set_status(BaseLineStatus.baseline_subtracted)
        self.plot()

    def plot_baseline(self):
        if self.profile.baseline is not None:
            self.baseline_plot.setData(self.profile.x, self.profile.baseline,
                                       pen=get_pen(width=4, color='red', style=Qt.DashDotLine))

    def _set_default_bounds(self):
        if self.x is None:
            self.profile.x_range = (0, 1)
        else:
            self.profile.x_range = (self.x.min(), self.x.max())

    def _update_bounds(self):
        self.profile.x_range = self._roi.getRegion()

    def _set_status(self, status: 'BaseLineStatus'):
        self._status = status
        self._baseline_setup_widget.set_status(status)

    def _on_calculate_baseline(self, params: dict):
        self.profile.set_parameters(**params)
        self._update_bounds()
        try:
            self.profile.update_baseline()
        except Exception as err:
            logger.exception(err)
            show_error('Failed calculating baseline. Change roi region or parameters and try again.',
                       error_title='Baseline calculation error')
            return
        self._set_status(BaseLineStatus.baseline_calculated)
        self.plot_baseline()

    def _on_subtracting_baseline(self):
        self.baseline_plot.clear()
        self._set_status(BaseLineStatus.baseline_subtracted)
        self.plot()

    def _on_restoring_data(self):
        self._set_status(BaseLineStatus.baseline_restored)
        self.plot_baseline()
        self.plot()

    def _on_closing_setup(self):
        self._baseline_setup_widget.hide()
        self._roi.hide()
        self.clear_baseline()
        if self._status != BaseLineStatus.baseline_subtracted:
            self._set_status(BaseLineStatus.no_baseline)
            self.profile.clear_baseline(clear_range=False)
        self.sigBackgroundChanged.emit()

    def clear_baseline(self):
        self.baseline_plot.clear()


class BaseLineStatus(Enum):
    no_baseline = 1
    baseline_calculated = 2
    baseline_restored = 3
    baseline_subtracted = 4


class BaseLineSetup(QWidget):
    calculate_signal = pyqtSignal(dict)
    subtract_signal = pyqtSignal()
    restore_signal = pyqtSignal()
    close_signal = pyqtSignal()

    def __init__(self, parent, status: BaseLineStatus, smoothness: float, asymmetry: float):
        super().__init__(parent=parent)
        self.setAttribute(Qt.WA_DeleteOnClose, False)
        self.setWindowFlag(Qt.WindowStaysOnTopHint)
        self.setWindowFlag(Qt.Window)
        self.setWindowTitle('Baseline setup')
        self.setWindowIcon(Icon('baseline'))
        self.setGeometry(0, 0, 400, 300)
        self._status = None
        self._init_ui(smoothness, asymmetry)
        self.set_status(status)
        center_widget(self)

    def _init_ui(self, smoothness: float, asymmetry: float):
        layout = QVBoxLayout(self)

        self.smoothness_slider = LabeledSlider('Smoothness', (1e1, 1e10), smoothness,
                                               self, Qt.Horizontal, decimals=3, scientific=True, log_scale=True)

        self.asymmetry_slider = LabeledSlider('Asymmetry', (0.0001, 1), asymmetry,
                                              self, Qt.Horizontal, decimals=3, scientific=True, log_scale=True)

        # self.save_params_box = QRadioButton('Save as default')
        # self.save_params_box.setChecked(True)
        self.calculate_button = QPushButton('Calculate baseline')
        self.calculate_button.clicked.connect(self.emit_calculate)
        self.subtract_button = QPushButton('Subtract baseline')
        self.subtract_button.clicked.connect(self.subtract_signal.emit)
        self.restore_button = QPushButton('Restore line')
        self.restore_button.clicked.connect(self.restore_signal.emit)

        layout.addWidget(self.smoothness_slider)
        layout.addWidget(self.asymmetry_slider)
        # layout.addWidget(self.save_params_box)
        layout.addWidget(self.calculate_button)
        layout.addWidget(self.subtract_button)
        layout.addWidget(self.restore_button)

    def set_status(self, status: BaseLineStatus):
        if status == BaseLineStatus.no_baseline:
            self.calculate_button.setEnabled(True)
            self.subtract_button.setEnabled(False)
            self.restore_button.setEnabled(False)
        elif status == BaseLineStatus.baseline_calculated:
            self.subtract_button.setEnabled(True)
            self.restore_button.setEnabled(False)
            self.calculate_button.setEnabled(True)
        elif status == BaseLineStatus.baseline_subtracted:
            self.subtract_button.setEnabled(False)
            self.restore_button.setEnabled(True)
            self.calculate_button.setEnabled(False)
        elif status == BaseLineStatus.baseline_restored:
            self.subtract_button.setEnabled(True)
            self.restore_button.setEnabled(False)
            self.calculate_button.setEnabled(True)
        else:
            logger.error(f'Unknown status {status}')
            return
        self._status = status

    def get_params_dict(self):
        return dict(smoothness=self.smoothness_slider.value,
                    asymmetry=self.asymmetry_slider.value)

    def set_parameters(self, params: dict):
        if 'smoothness' in params:
            self.smoothness_slider.set_value(params['smoothness'], True)
        if 'asymmetry' in params:
            self.asymmetry_slider.set_value(params['asymmetry'], True)

    def emit_calculate(self):
        self.calculate_signal.emit(self.get_params_dict())

    def closeEvent(self, a0) -> None:
        # if self.save_params_box.isChecked():
        #     save_config(self.NAME, self.get_params_dict())
        self.close_signal.emit()
        super().closeEvent(a0)



from h5py import Group

from .object_file_manager import _ObjectFileManager
from ..rois.roi_data import RoiData


class _ReadRoiData(_ObjectFileManager):
    NAME = 'roi_data'

    @staticmethod
    def _get_h5(h5group: Group, key):
        if 'roi_data' in h5group.keys():
            return RoiData.from_array(h5group['roi_data'][()])

    @staticmethod
    def _set_h5(h5group: Group, key, value: RoiData):
        if 'roi_data' in h5group.keys():
            del h5group['roi_data']
        h5group.create_dataset('roi_data', data=value.to_array())

    @staticmethod
    def _del_h5(h5group: Group, key):
        if 'roi_data' in h5group.keys():
            del h5group['roi_data']



# -*- coding: utf-8 -*-

from .basic_profile import BasicProfile


class AngularProfile(BasicProfile):

    def __init__(self, image_holder, parent=None):
        self.image_holder = image_holder
        super().__init__(parent)

    def update_data_from_source(self, key: int = None):

        image_holder = self.image_holder
        profile = image_holder.get_angular_profile(key)
        if profile is None:
            return
        phi_axis = image_holder.geometry.phi_axis
        if phi_axis is not None and phi_axis.size == profile.size:
            self.set_data(profile, phi_axis)



from enum import Enum
from abc import abstractmethod
from typing import Tuple, Dict, Callable

import numpy as np

from .utils import Roi, _update_bounds
from .background import Background

__all__ = ['FittingType', 'FittingFunction', 'FITTING_FUNCTIONS', 'Gaussian', 'Lorentzian']


class FittingType(Enum):
    gaussian = 'Gaussian'
    lorentzian = 'Lorentzian'


class FittingFunction(object):
    NAME: str = ''
    PARAM_NAMES: tuple = ()
    NUM: int = 0
    TYPE: FittingType = None
    is_default: bool = True

    def __init__(self, *args, **kwargs):
        pass

    def get_func(self, background: Background) -> Callable:
        def func(x: np.ndarray, *params):
            return self.__call__(x, background, *params)

        return func

    def __call__(self, x: np.ndarray, background: Background, *params) -> np.ndarray:
        return self.func(x, *params[:self.NUM]) + background(x, *params[self.NUM:])

    @staticmethod
    @abstractmethod
    def func(x: np.ndarray, *params):
        pass

    @staticmethod
    @abstractmethod
    def set_roi_from_params(roi: Roi, params: list):
        pass

    @staticmethod
    @abstractmethod
    def set_params_from_roi(roi: Roi, params: list):
        pass

    @staticmethod
    @abstractmethod
    def _bounds(x: np.ndarray, y: np.ndarray, roi: Roi, background: Background):
        pass

    @classmethod
    def bounds(cls, x: np.ndarray, y: np.ndarray, roi: Roi, background: Background, params_from_roi: bool = False):
        if params_from_roi:
            return _update_bounds(roi, list(cls.PARAM_NAMES) + list(background.PARAM_NAMES),
                                  *cls._bounds(x, y, roi, background))
        else:
            return cls._bounds(x, y, roi, background)


class Gaussian(FittingFunction):
    NAME: str = 'Gaussian'
    PARAM_NAMES: tuple = ('peak height', 'radius', 'width')
    NUM = 3
    TYPE = FittingType.gaussian

    @staticmethod
    def func(x: np.ndarray, *params):
        amp, mu, sigma, *_ = params
        return amp * np.exp(- 2 * (x - mu) ** 2 / sigma ** 2)

    @staticmethod
    def set_roi_from_params(roi: Roi, params: list):
        roi.radius = params[1]
        roi.width = params[2]

    @staticmethod
    def _bounds(x: np.ndarray, y: np.ndarray, roi: Roi, background: Background):
        init_b, upper_b, lower_b = background.bounds(x, y, roi)
        amp, amp_max, amp_min = background.amp_bounds(x, y, init_b)

        return ([amp, roi.radius, roi.width] + init_b,
                [amp_max, roi.radius + roi.width / 2, roi.width * 2] + upper_b,
                [amp_min, roi.radius - roi.width / 2, 0] + lower_b)


class Lorentzian(Gaussian):
    NAME: str = 'Lorentzian'
    PARAM_NAMES: tuple = ('peak height', 'radius', 'width')
    TYPE = FittingType.lorentzian

    @staticmethod
    def func(x: np.ndarray, *params):
        amp, mu, sigma, *_ = params
        w = (sigma / 2) ** 2
        return amp * w / (w + (x - mu) ** 2)


FITTING_FUNCTIONS: Dict[FittingType, FittingFunction.__class__] = {
    FittingType.gaussian: Gaussian,
    FittingType.lorentzian: Lorentzian,
}



import pickle
import logging
from pathlib import Path
from typing import List


class _GlobalConfigManager(object):
    config_folder: Path = Path(__file__).parents[2] / 'config'
    log = logging.getLogger(__name__)

    def __init__(self):
        if not self.config_folder.is_dir():
            self.config_folder.mkdir(parents=False)
        self._saved_projects_path = str((self.config_folder / 'recent_projects').resolve())

    def get_project_paths(self) -> List[Path]:
        try:
            with open(self._saved_projects_path, 'rb') as f:
                return pickle.load(f)
        except FileNotFoundError:
            return []

    def add_project_path(self, path: Path):
        paths: List[Path] = self.get_project_paths()
        paths.append(path)
        with open(self._saved_projects_path, 'wb') as f:
            pickle.dump(paths, f)

    def update_project_paths(self, paths: List[Path]):
        with open(self._saved_projects_path, 'wb') as f:
            pickle.dump(paths, f)

    def __getitem__(self, item):

        try:
            with open(str((self.config_folder / item).resolve()), 'rb') as f:
                return pickle.load(f)
        except FileNotFoundError:
            return
        except Exception as err:
            self.log.exception(err)
            return

    def __setitem__(self, key, value):

        with open(str((self.config_folder / key).resolve()), 'wb') as f:
            pickle.dump(value, f)

    def __delitem__(self, key):

        path = self.config_folder / key
        if path.is_file():
            path.unlink()



# -*- coding: utf-8 -*-

import os
import logging
from typing import List
from pathlib import Path

from PyQt5.QtWidgets import (QWidget, QPushButton, QGridLayout,
                             QVBoxLayout, QFileDialog, QLabel,
                             QLineEdit, QScrollArea, QSizePolicy)
from PyQt5.QtCore import Qt, pyqtSignal, pyqtSlot

from .tools import center_widget, show_error, Icon, color_animation, get_folder_filepath
from .basic_widgets import Label
from ..__version import __version__


class InitWindow(QWidget):
    sigOpenProject = pyqtSignal(object)
    sigExit = pyqtSignal()

    log = logging.getLogger(__name__)

    WIDTH = 600
    HEIGHT = 400

    def __init__(self,
                 recent_projects: List[Path] = None,
                 is_updated: bool = False):
        flags = Qt.WindowFlags()
        flags |= Qt.FramelessWindowHint
        super().__init__(flags=flags)

        self.setAttribute(Qt.WA_DeleteOnClose)
        self.setWindowTitle('GIWAXS analysis')
        self.setWindowIcon(Icon('window_icon'))
        self.project_name: str = 'untitled'
        self.project_path: Path = Path('~/GIWAXS projects').expanduser().resolve() / self.project_name

        self.__init_ui(recent_projects, is_updated)
        self.setFixedWidth(self.WIDTH)
        self.setFixedHeight(self.HEIGHT)
        center_widget(self)
        self.show()

    def __init_ui(self, recent_projects: List[Path] or None, is_updated: bool = False):
        layout = QGridLayout(self)
        layout.setContentsMargins(40, 20, 40, 40)
        self.file_line = QLineEdit(str(self.project_path), self)
        self.file_line.textEdited.connect(self._on_text_editing)

        browse_action = self.file_line.addAction(Icon('folder'), QLineEdit.LeadingPosition)

        browse_action.triggered.connect(self._new_project_dialog)
        self.create_button = QPushButton(f'Create project "{self.project_name}"', self)
        self.create_button.clicked.connect(self._create)

        if is_updated:
            title = f'GIWAXS analysis. Updated to the latest version {__version__}!'
        else:
            title = f'GIWAXS analysis (version {__version__})'

        layout.addWidget(Label(title, self, 11, True), 0, 0, 1, 3, alignment=Qt.AlignHCenter)
        layout.addWidget(Label('New project', self, 9.5), 1, 0, 1, 3)
        layout.addWidget(self.file_line, 2, 0, 1, 2)
        layout.addWidget(self.create_button, 2, 2)

        if recent_projects:
            layout.addWidget(self._init_recent_projects_area(recent_projects), 3, 0, 1, 3)

        layout.addLayout(self._init_exit_button(), 4 if recent_projects else 3, 0, 1, 3)

    def _init_exit_button(self):
        e_layout = QVBoxLayout()
        exit_button = QPushButton('Exit')
        exit_button.clicked.connect(self.sigExit)
        e_layout.addWidget(QLabel(''))
        e_layout.addWidget(exit_button)
        return e_layout

    def _init_recent_projects_area(self, recent_projects):
        q_scroll_area = QScrollArea(self)
        q_scroll_area.setWidgetResizable(True)
        q_scroll_area.setGeometry(0, 0, self.WIDTH, self.HEIGHT // 2)

        projects_list_widget = QWidget(self)
        projects_list_widget.setSizePolicy(
            QSizePolicy.Expanding, QSizePolicy.Expanding
        )

        scroll_layout = QVBoxLayout(projects_list_widget)
        scroll_layout.addWidget(QLabel('', self))
        scroll_layout.addWidget(Label('Recent projects', self, 9.5))
        for path in recent_projects:
            btn = QPushButton(path.name)
            btn.clicked.connect(
                lambda *x, p=path: self.sigOpenProject.emit(p))
            scroll_layout.addWidget(btn)
        q_scroll_area.setWidget(projects_list_widget)

        return q_scroll_area

    @pyqtSlot(str, name='onTextEditing')
    def _on_text_editing(self, path: str):
        self.project_name = path.split(os.sep)[-1]
        self.project_path = self.project_path.parent / self.project_name
        self.create_button.setText(f'Create project "{self.project_name}"')

    @pyqtSlot(name='NewProjectDialog')
    def _new_project_dialog(self):
        folder = QFileDialog.getExistingDirectory(
            self, 'New project folder', options=
            QFileDialog.ShowDirsOnly | QFileDialog.DontUseNativeDialog
        )
        if not folder:
            return
        folder = Path(folder).resolve()
        if folder.is_dir():
            self.project_path = folder / self.project_name
            self.file_line.setText(str(self.project_path))
        else:
            color_animation(self.file_line)
            show_error('Invalid folder. Please, select existing folder',
                       error_title='Wrong path')

    def _create(self):
        try:
            self.project_path.mkdir(parents=True, exist_ok=False)
            self.sigOpenProject.emit(self.project_path.resolve())
        except FileExistsError:
            show_error(f'The folder {str(self.project_path.resolve())} already exists', error_title='Folder exists')
            color_animation(self.file_line)
        except PermissionError:
            folder = get_folder_filepath(self, 'Permission denied, please, create a project folder manually',
                                         directory=str(Path('~').expanduser().resolve()))
            if folder and folder.is_dir():
                self.sigOpenProject.emit(folder.resolve())



import logging
from pathlib import Path
from typing import List

from h5py import File, Group

from PyQt5.QtCore import pyqtSignal, QObject

from .keys import (FolderKey, FolderH5Key, FolderPathKey, RemoveWeakrefs,
                   ImageKey, ImageH5Key, ImagePathKey, InvalidKey,
                   PROJECT_KEY, IMAGE_PROJECT_KEY, GLOB_IMAGE_FORMATS)

from .project_structure import ProjectStructure, ProjectRootKey
from .read_images import _ReadImage, _ReadNpy
from .read_polar_images import _ReadPolarImage
from .read_geometry import _ReadGeometry
from .read_roi_data import _ReadRoiData
from .read_meta_roi import _ReadMetaData
from .read_radial_profile import _ReadRadialProfile
from .config_manager import _GlobalConfigManager
from .read_fits import _ReadFits


class FileManager(QObject):
    sigActiveImageChanged = pyqtSignal(object)
    sigProjectClosed = pyqtSignal()
    sigProjectIsClosing = pyqtSignal()
    sigProjectOpened = pyqtSignal()
    sigNewFolder = pyqtSignal(object)
    sigNewFile = pyqtSignal(object)

    log = logging.getLogger(__name__)

    def __init__(self):
        QObject.__init__(self)
        self.config: _GlobalConfigManager = _GlobalConfigManager()
        self.recent_projects: List[Path] = self.config.get_project_paths()
        self._project_folder: Path or None = None
        self._project_structure: ProjectStructure = ProjectStructure()
        self.project_name: str = None
        self._current_key: ImageKey = None

        self.recent_projects = [p for p in self.recent_projects if p.is_dir()]

        # self.open_latest_available_project()

    def open_latest_available_project(self):
        self.close_project()
        while self.recent_projects:
            path = self.recent_projects.pop(-1)
            if path.is_dir():
                try:
                    self.open_project(path)
                    return
                except Exception as err:
                    self.log.exception(err)

    @property
    def project_opened(self):
        return self._project_structure.project_opened

    def change_image(self, key: ImageKey):
        if key != self._current_key:
            self._current_key = key
            self.sigActiveImageChanged.emit(key)

    def open_project(self, path: Path):
        if path == self._project_folder:
            return
        self.close_project()
        self._init_project(path)
        self.sigProjectOpened.emit()

    def add_root_path_to_project(self, path: Path):
        if not self.project_opened:
            return
        key = self._project_structure.root.add_path(path)
        if isinstance(key, ImageKey):
            self.sigNewFile.emit(key)
        elif isinstance(key, FolderKey):
            self.sigNewFolder.emit(key)

    def remove_key(self, key):
        self._project_structure.root.remove_key(key)
        if isinstance(key, ImageKey):
            self._delete_image_data(key)
            if key.parent:
                key.parent.remove_image(key)
        else:
            self._delete_folder(key)
        if self._current_key in key:
            self.change_image(None)

    def _delete_image_data(self, key: ImageKey):
        del self.geometries[key]
        del self.rois_data[key]
        del self.images[key]
        del self.polar_images[key]

        # TODO find all fits by image key (?)

    def _delete_folder(self, key: FolderKey):
        for folder in key.folder_children:
            self._delete_folder(folder)
        for image in key.image_children:
            self._delete_image_data(image)

    def close_project(self):
        if self.project_opened:
            self.sigProjectIsClosing.emit()
            self._project_structure.save_and_close()
            if self._project_folder not in self.recent_projects:
                self.recent_projects.append(self._project_folder)
            self._project_folder = None
            self.project_name = None
            self.sigProjectClosed.emit()

    def _init_project(self, path: Path):
        self._project_folder = path
        self._project_folder.mkdir(parents=False, exist_ok=True)
        self._project_structure.open_project(path)

        self.images: _ReadImage = _ReadImage(self._project_structure)
        self.geometries: _ReadGeometry = _ReadGeometry(self._project_structure)
        self.polar_images: _ReadPolarImage = _ReadPolarImage(self._project_structure)
        self.rois_data: _ReadRoiData = _ReadRoiData(self._project_structure)
        self.fits: _ReadFits = _ReadFits(self._project_structure)
        self.profiles: _ReadRadialProfile = _ReadRadialProfile(self._project_structure)
        self.project_name = self._project_folder.name

        for key in self._project_structure.root.folder_children:
            self.sigNewFolder.emit(key)
        for key in self._project_structure.root.image_children:
            self.sigNewFile.emit(key)

    def save_as_h5(self, h5path: Path):
        with File(str(h5path.resolve()), 'w') as f:
            f.attrs[PROJECT_KEY] = True
        self._save_folder_as_h5(h5path, self._project_structure.root)

    def _save_folder_as_h5(self, h5path: Path, folder_key: FolderKey):
        if folder_key.images_num > 0:
            with File(str(h5path.resolve()), 'a') as f:
                group = f.create_group(_give_h5_name(f, folder_key.name))

                for image in folder_key.image_children:
                    self._save_image_as_h5(group, image)
        for folder in folder_key.folder_children:
            self._save_folder_as_h5(h5path, folder)

    def _save_image_as_h5(self, h5group: Group, image_key: ImageKey):
        img_group = h5group.create_group(_give_h5_name(h5group, image_key.name))
        img_group.attrs[IMAGE_PROJECT_KEY] = True
        self.images._set_h5(img_group, image_key, image_key.get_image())
        roi_data = self.rois_data[image_key]
        if roi_data:
            self.rois_data._set_h5(img_group, image_key, roi_data)
        geometry = self.geometries[image_key] or self.geometries.default[image_key]
        if geometry:
            self.geometries._set_h5(img_group, image_key, geometry)

    def __len__(self):
        return len(self.paths)

    def close(self):
        self.close_project()
        self.config.update_project_paths(self.recent_projects)


def _give_h5_name(h5group: Group, name):
    if name not in h5group.keys():
        return name
    num = 0
    while True:
        new_name = '_'.join((name, str(num)))
        if new_name not in h5group.keys():
            return new_name
        num += 1



from pathlib import Path
import logging

from PyQt5.QtWidgets import (QMainWindow, QWidget, QSizePolicy,
                             QApplication, QShortcut, QMessageBox,
                             QFileDialog)
from PyQt5.QtCore import (Qt, pyqtSlot, pyqtSignal,
                          QObject, QThreadPool)
from PyQt5.QtGui import QKeySequence

import qdarkstyle
import qdarkgraystyle

from ..app import App
from ..app.update import CheckVersionMessage, check_outdated, update_package
from ..app.utils import Worker
from ..__version import __version__

from .dock_area import AppDockArea
from .basic_widgets import ToolBar
from .tools import Icon, get_image_filepath, get_folder_filepath, save_file_dialog

from .init_window import InitWindow
from .debug_widgets import DebugWindow
from .exception_message import UncaughtHook
from .notifications import (PopUpWrapper, NotificationTypes,
                            CheckingVersion, AskRestart, UpdatingProgram,
                            UpdateFailed, get_check_result_notification)


class GIWAXSMainController(QObject):
    EXIT_CODE_REBOOT: int = -123456789

    log = logging.getLogger(__name__)

    def __init__(self):
        super().__init__()
        self.app = App()

        if self.app.fm.config['is_updated']:
            self.is_updated = True
            del self.app.fm.config['is_updated']
        else:
            self.is_updated = False

        self.init_window = None
        self.main_window = None

        if self.log.level <= logging.DEBUG:
            self.debug_window = DebugWindow()
        self.exception_hook = UncaughtHook()

        self.log.info(f'{"*" * 10}')
        self.log.info(f'Starting GIWAXS analysis {__version__}!')
        self.log.info(f'{"*" * 10}')
        self.set_style(self.app.fm.config['style'] or 'Gray Dark')
        self.open_init_window()

    def open_main_window(self):
        self.main_window = GIWAXSMainWindow(self.is_updated)

        self.main_window.sigCloseApp.connect(self.close_app)
        self.main_window.sigRestartApp.connect(self.restart_app)
        self.main_window.sigRestartAfterUpdate.connect(self.restart_after_update)
        self.main_window.sigOpenProject.connect(self.open_new_project)
        self.main_window.sigCloseProject.connect(self.close_project)
        self.main_window.sigSetStyle.connect(self.set_style)

    def open_init_window(self):
        self.init_window = InitWindow(self.app.fm.recent_projects, self.is_updated)
        self.init_window.sigOpenProject.connect(self.open_new_project)
        self.init_window.sigExit.connect(self.close_app)

    @pyqtSlot(name='closeProject')
    def close_project(self):
        self.app.fm.close_project()
        self.main_window.hide()
        self.open_init_window()

    @pyqtSlot(object, name='NewProject')
    def open_new_project(self, path: Path):
        if self.init_window:
            self.init_window.close()
            self.init_window = None
        if not self.main_window:
            self.open_main_window()
        else:
            self.main_window.show()
        self.app.fm.open_project(path)

    @pyqtSlot(name='restartAppAfterUpdate')
    def restart_after_update(self):
        self.app.fm.config['is_updated'] = True
        self.restart_app()

    @pyqtSlot(name='restartApp')
    def restart_app(self):
        self.app.close()
        q_app = QApplication.instance()
        App._instance = None
        q_app.exit(self.EXIT_CODE_REBOOT)

    @pyqtSlot(name='closeApp')
    def close_app(self):
        self.app.close()
        q_app = QApplication.instance()
        App._instance = None
        q_app.exit(0)

    @pyqtSlot(str, name='setStyle')
    def set_style(self, name: str = 'Gray Dark'):
        q_app = QApplication.instance()
        if not q_app:
            raise RuntimeError('No running application found.')
        if name == 'Dark':
            q_app.setStyleSheet('')
            q_app.setStyleSheet(qdarkstyle.load_stylesheet_pyqt5())
        elif name == 'Gray Dark':
            q_app.setStyleSheet('')
            q_app.setStyleSheet(qdarkgraystyle.load_stylesheet_pyqt5())
        else:
            return

        self.app.fm.config['style'] = name


class GIWAXSMainWindow(QMainWindow):
    _MinimumSize = (500, 500)

    sigCloseApp = pyqtSignal()
    sigOpenProject = pyqtSignal(object)
    sigCloseProject = pyqtSignal()
    sigRestartApp = pyqtSignal()
    sigRestartAfterUpdate = pyqtSignal()
    sigSetStyle = pyqtSignal(str)

    def __init__(self, is_updated: bool, parent=None):
        super(GIWAXSMainWindow, self).__init__(parent=parent)
        self.__closing: bool = False
        self.app = App()
        self._init_toolbar()
        self._init_shortcuts()
        self._init_menubar()
        self.popups = PopUpWrapper(self)

        self.q_thead_pool = QThreadPool(self)

        self.dock_area = AppDockArea(self)
        self.app.fm.sigProjectClosed.connect(self.update_window_title)
        self.app.fm.sigProjectOpened.connect(self.update_window_title)

        self.setCentralWidget(self.dock_area)
        self.update_window_title()
        self.setWindowIcon(Icon('window_icon'))
        self.setMinimumSize(*self._MinimumSize)
        self.setWindowState(Qt.WindowMaximized)
        self.show()

        if not is_updated:
            self.check_update()
        else:
            self.popups.add_notification(get_check_result_notification(CheckVersionMessage.latest_version_installed))

    def check_update(self):
        self.popups.add_notification(CheckingVersion(), NotificationTypes.checking_version)
        worker = Worker(check_outdated, version=__version__)
        worker.autoDelete()
        worker.signals.result.connect(self._version_checked)
        self.q_thead_pool.start(worker)

    def update_package(self, target_version: str = ''):
        worker = Worker(update_package, version=target_version)
        worker.autoDelete()
        worker.signals.result.connect(self._package_updated)
        self.q_thead_pool.start(worker)
        self.popups.add_notification(UpdatingProgram(target_version), NotificationTypes.updating_program)

    @pyqtSlot(object, name='versionChecked')
    def _version_checked(self, res: CheckVersionMessage):
        self.popups.remove_by_name(NotificationTypes.checking_version)
        self.popups.add_notification(get_check_result_notification(res), NotificationTypes.check_result)

        if res.value == CheckVersionMessage.new_version_available.value:
            try:
                self.update_package(res.version)
            except AttributeError:
                return

    @pyqtSlot(object, name='packageUpdated')
    def _package_updated(self, is_updated: bool):
        self.popups.remove_by_name(NotificationTypes.updating_program)
        if is_updated:
            ask_restart = AskRestart()
            ask_restart.sigRestart.connect(self.sigRestartAfterUpdate)
            self.popups.add_notification(ask_restart)
        else:
            self.popups.add_notification(UpdateFailed())

    def update_window_title(self):
        if self.app.fm.project_opened:
            self.setWindowTitle(f'{self.app.fm.project_name} - GIWAXS analysis')
        else:
            self.setWindowTitle('GIWAXS analysis')

    @pyqtSlot(name='NewProjectDialog')
    def _new_project_dialog(self):
        folder = QFileDialog.getExistingDirectory(self, 'New project folder')
        if not folder:
            return
        folder = Path(folder)
        if folder.is_dir():
            self.sigOpenProject.emit(folder)

    @pyqtSlot(name='addExSitu')
    def _add_file_dialog(self):
        filepath = get_image_filepath(self)
        if filepath:
            self.app.fm.add_root_path_to_project(filepath)

    def _save_as_h5(self):
        path = save_file_dialog(self, title='Save project')
        if path:
            self.app.fm.save_as_h5(path)

    def _add_folder_dialog(self):
        filepath = get_folder_filepath(self, message='Choose directory containing images or h5 files')
        if filepath:
            self.app.fm.add_root_path_to_project(filepath)

    def _init_menubar(self):
        self.menubar = self.menuBar()

        # File menu

        self.file_menu = self.menubar.addMenu('File')
        self.file_menu.addAction('New project', self._new_project_dialog)
        recent_projects_menu = self.file_menu.addMenu('Recent projects')

        for project_path in self.app.fm.recent_projects:
            recent_projects_menu.addAction(
                project_path.name, lambda *x, p=project_path: self.sigOpenProject.emit(p))

        # Save

        self.save_menu = self.file_menu.addMenu('Save project as ...')
        self.save_menu.addAction('Save as h5 file', self._save_as_h5)

        # Close project

        self.file_menu.addAction('Close project', lambda *x: self.sigCloseProject.emit())

        # Restart

        self.file_menu.addAction('Restart', lambda *x: self.restart())

        # Data menu

        self.data_menu = self.menubar.addMenu('Data')
        self.add_data_menu = self.data_menu.addMenu('Add data')

        self.add_data_menu.addAction('Add file', self._add_file_dialog)
        self.add_data_menu.addAction('Add folder', self._add_folder_dialog)

        # Preferences menu

        self.preferences = self.menubar.addMenu('Preferences')
        self.themes_menu = self.preferences.addMenu('Themes')
        themes = ['Dark', 'Gray Dark']
        # themes = CSS.list_css() + QStyleFactory.keys()
        for theme in themes:
            theme_action = self.themes_menu.addAction(theme)
            theme_action.triggered.connect(lambda *x, t=theme: self.sigSetStyle.emit(t))

    def _init_shortcuts(self):
        self.copy_shortcut = QShortcut(QKeySequence("Ctrl+C"), self)
        self.copy_shortcut.activated.connect(lambda *x: self.app.roi_dict.copy_rois('selected'))

        self.paste_shortcut = QShortcut(QKeySequence("Ctrl+V"), self)
        self.paste_shortcut.activated.connect(lambda *x: self.app.roi_dict.paste_rois())

        self.select_shortcut = QShortcut(QKeySequence("Ctrl+A"), self)
        self.select_shortcut.activated.connect(lambda *x: self.app.roi_dict.select_all())

        self.delete_shortcut = QShortcut(QKeySequence('Del'), self)
        self.delete_shortcut.activated.connect(lambda *x: self.app.roi_dict.delete_selected_roi())

        self.fit_shortcut = QShortcut(QKeySequence('Ctrl+F'), self)
        self.fit_shortcut.activated.connect(lambda *x: self.app.roi_dict.open_fit_rois(True))

        def raise_err(*_):
            raise ValueError('Congratulations, you found an error combination used for testing! '
                             'Your project will be deleted in 3 seconds...')

        self.raise_shortcut = QShortcut(QKeySequence('Ctrl+R'), self)
        self.raise_shortcut.activated.connect(raise_err)

    def _init_toolbar(self):

        docks_toolbar = ToolBar('Docks', self)
        self.addToolBar(docks_toolbar)

        control_widget = docks_toolbar.addAction(Icon('folder'), 'File Manager')
        control_widget.triggered.connect(lambda: self.dock_area.show_hide_docks('file_widget'))

        radial_profile = docks_toolbar.addAction(Icon('radial_profile'), 'Radial profile')
        radial_profile.triggered.connect(lambda: self.dock_area.show_hide_docks('radial_profile'))

        angular_profile = docks_toolbar.addAction(Icon('angular_profile'), 'Angular profile')
        angular_profile.triggered.connect(lambda: self.dock_area.show_hide_docks('angular_profile'))

        interpolation = docks_toolbar.addAction(Icon('interpolate'), 'Polar Viewer')
        interpolation.triggered.connect(lambda: self.dock_area.show_hide_docks('polar'))

        self.gen_toolbar = ToolBar('General')
        self.addToolBar(self.gen_toolbar)
        spacer_widget = QWidget()
        spacer_widget.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Preferred)
        spacer_widget.setVisible(True)
        self.gen_toolbar.addWidget(spacer_widget)

        self.fullscreen_action = self.gen_toolbar.addAction(Icon('tofullscreen'), 'Full screen')
        self.fullscreen_action.triggered.connect(self._on_fullscreen_changed)

    def _on_fullscreen_changed(self):
        if self.isFullScreen():
            self.setWindowState(Qt.WindowMaximized)
            self.fullscreen_action.setIcon(Icon('tofullscreen'))
        else:
            self.setWindowState(Qt.WindowFullScreen)
            self.fullscreen_action.setIcon(Icon('fromfullscreen'))

    @pyqtSlot(name='restartApp')
    def restart(self):
        self.__closing = True
        self.sigRestartApp.emit()

    def closeEvent(self, a0) -> None:
        if self.__closing:
            super().closeEvent(a0)
        else:
            reply = QMessageBox.question(self, 'Message',
                                         "Are you sure to quit? The project will be saved.",
                                         QMessageBox.Yes |
                                         QMessageBox.No, QMessageBox.No)

            if reply == QMessageBox.Yes:
                a0.accept()
                self.__closing = True
                self.sigCloseApp.emit()
            else:
                a0.ignore()



from setuptools import setup, find_packages
from pathlib import Path
import re

PACKAGE_NAME = 'giwaxs_gui'


def read(filename: str):
    with open(Path(__file__).parent / filename, mode='r', encoding='utf-8') as f:
        return f.read()


def get_version():
    version_file = f'{PACKAGE_NAME}/__version.py'
    with open(version_file, 'r') as f:
        file_str = f.read()

    mo = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]", file_str, re.M)
    if mo:
        return mo.group(1)
    else:
        raise RuntimeError("Unable to find version string in %s." % (version_file,))


setup(
    name=PACKAGE_NAME,
    packages=find_packages(),
    version=get_version(),
    author='Vladimir Starostin',
    author_email='vladimir.starostin@uni-tuebingen.de',
    description='A GUI program for GIWAXS images analysis',
    long_description=read('README.md'),
    long_description_content_type='text/markdown',
    license='GPLv3',
    python_requires='>=3.7.2',
    entry_points={
        'gui_scripts': [
            'giwaxs_gui = giwaxs_gui:main',
        ],
        'console_scripts': [
            'giwaxs_gui_update = giwaxs_gui.app.update:giwaxs_gui_update'],
    },
    install_requires=[
        'numpy>=1.18.1',
        'opencv-python>=4.0.0.0',
        'scipy>=1.4.1',
        'h5py>=2.10.0',
        'PyQt5==5.13.0',
        'pyqtgraph==0.11.0rc0',
        'read_edf',
        'Pillow',
        'requests',
        'qdarkstyle',
        'qdarkgraystyle'
    ],
    include_package_data=True,
    keywords='xray python giwaxs scientific-analysis',
    url='https://pypi.org/project/giwaxs-gui',
)



# -*- coding: utf-8 -*-

from ..basic_widgets import PlotBC
from ...app.app import App
from ..roi_widgets.abstract_roi_holder import AbstractRoiHolder
from ..roi_widgets.roi_1d_widget import Roi1DAngular


class AngularProfileWidget(AbstractRoiHolder, PlotBC):
    def __init__(self, parent=None):
        AbstractRoiHolder.__init__(self, 'AngularProfile')
        PlotBC.__init__(self, App().angular_profile, parent)
        self._selected_key = None

        # self.image_view.plot_item.setTitle('Angular Profile')
        self.image_view.plot_item.getAxis('bottom').setLabel('&Phi;', color='white', font_size='large')
        self.image_view.plot_item.getAxis('left').setLabel('Intensity', color='white', font_size='large')

    def _init_connect(self):
        App().image_holder.sigPolarImageChanged.connect(self._update)
        App().image_holder.sigEmptyImage.connect(self.clear_plot)
        self._roi_dict.sig_roi_created.connect(self._update)
        self._roi_dict.sig_roi_deleted.connect(self._update)
        self._roi_dict.sig_roi_moved.connect(self._update)
        self._roi_dict.sig_roi_moved.connect(self._move_rois)
        self._roi_dict.sig_selected.connect(self._update)
        self._roi_dict.sig_one_selected.connect(self._update)
        self._roi_dict.sig_fixed.connect(self._fix_rois)
        self._roi_dict.sig_unfixed.connect(self._unfix_rois)

    def _update(self):
        selected_rois = self._roi_dict.selected_rois

        if len(selected_rois) == 1:
            selected_roi = selected_rois[0]
            key = selected_roi.key
            if key == self._selected_key:
                self.profile.update()
            else:
                self._clear()
                self._selected_key = key
                self._create_roi((key, ))
                self.profile.update()
        else:
            self._clear()

    def _clear(self):
        if self._selected_key is not None:
            try:
                self._delete_roi_widget(self._roi_widgets.pop(self._selected_key))
            except KeyError:
                pass
            self._selected_key = None
        self.clear_plot()

    def _make_roi_widget(self, roi):
        roi_widget = Roi1DAngular(roi)
        self.image_view.plot_item.addItem(roi_widget)
        return roi_widget

    def _delete_roi_widget(self, roi_widget):
        self.image_view.plot_item.removeItem(roi_widget)



import subprocess
import sys
import logging
import argparse
import platform
import os


def update_package(version: str = '', num_of_attempts: int = 2) -> bool:
    logger = logging.getLogger(__name__)
    logger.info(f'Updating the package to {version}...')

    try:
        if platform.system() == 'Windows':
            subprocess.check_call(['giwaxs_gui_update', f'--version={version}',
                                   f'--num_of_attempts={num_of_attempts}'],
                                  creationflags=subprocess.CREATE_NO_WINDOW)
        else:
            subprocess.check_call(['nohup', 'giwaxs_gui_update', f'--version={version}',
                                   f'--num_of_attempts={num_of_attempts}'],
                                  shell=False, stdout=None, stderr=None, preexec_fn=os.setpgrp)
        return True

    except (subprocess.CalledProcessError, FileNotFoundError) as err:
        logger.exception(err)
        return False


def giwaxs_gui_update() -> int:
    parser = argparse.ArgumentParser(description='Update giwaxs_gui package.')

    parser.add_argument('--version', type=str, default='', help='the target version')
    parser.add_argument('--num_of_attempts', type=int, default=2, help='number of attempts to update')
    parser.add_argument('--package', type=str, default='giwaxs_gui', help='the target package')

    args = parser.parse_args()
    num_of_attempts: int = args.num_of_attempts
    package: str = args.package
    version: str = args.version

    if version:
        package = f'{package}=={version}'

    for i in range(num_of_attempts):
        try:
            subprocess.check_call([sys.executable, '-m', 'pip', 'install', '--user', '--upgrade', package])
            return 0

        except subprocess.CalledProcessError:
            continue

    for i in range(num_of_attempts):
        try:
            subprocess.check_call([sys.executable, '-m', 'pip', 'install', '--upgrade', package])
            return 0

        except subprocess.CalledProcessError:
            continue

    return 1


if __name__ == '__main__':
    giwaxs_gui_update()



# -*- coding: utf-8 -*-
from abc import abstractmethod

from PyQt5.QtCore import QObject, pyqtSignal, Qt, QPointF
from pyqtgraph import ViewBox

from ...app.rois.roi import Roi


class MouseEventsController(QObject):
    def __init__(self, parent=None):
        super().__init__(parent)

    def mouseMoveEvent(self, ev) -> bool:
        return True

    def mouseReleaseEvent(self, ev) -> bool:
        return True

    def mousePressEvent(self, ev) -> bool:
        return True


class ViewBoxMouseEvents(MouseEventsController):
    def __init__(self, view_box: ViewBox = None, parent=None):
        super().__init__(parent)

        self._view_box = None
        self._press_func = None
        self._release_func = None
        self._move_func = None

        if view_box:
            self.attach_to_view_box(view_box)

    def attach_to_view_box(self, view_box: ViewBox):
        self._view_box = view_box
        self._press_func = view_box.mousePressEvent
        self._release_func = view_box.mouseReleaseEvent
        self._move_func = view_box.mouseMoveEvent

        def mouseMoveEvent(ev):
            if self.mouseMoveEvent(ev):
                self._move_func(ev)

        def mouseReleaseEvent(ev):
            if self.mouseReleaseEvent(ev):
                self._release_func(ev)

        def mousePressEvent(ev):
            if self.mousePressEvent(ev):
                self._press_func(ev)

        self._view_box.mousePressEvent = mousePressEvent
        self._view_box.mouseReleaseEvent = mouseReleaseEvent
        self._view_box.mouseMoveEvent = mouseMoveEvent

    def detach_from_view_box(self):
        if self._view_box:
            self._view_box.mousePressEvent = self._press_func
            self._view_box.mouseReleaseEvent = self._release_func
            self._view_box.mouseMoveEvent = self._move_func

        self._view_box = None
        self._press_func = None
        self._release_func = None
        self._move_func = None


class DrawRoiController(ViewBoxMouseEvents):
    sigCreateRoi = pyqtSignal(object)
    sigMoveRoi = pyqtSignal(int, str)

    def __init__(self, view_box=None, parent=None):
        super().__init__(view_box, parent)
        self._init_point: QPointF = None
        self._roi: Roi = None
        self._pressed: bool = False

    def mousePressEvent(self, ev) -> bool:
        if ev.button() == Qt.LeftButton and ev.modifiers() == Qt.ControlModifier:
            self._init_point = self._view_box.mapToView(ev.pos())
            self._roi = self._init_roi()
            self._update_roi(self._init_point)
            self.sigCreateRoi.emit(self._roi)
            self._pressed = True
            ev.accept()
            return False
        return True

    def mouseMoveEvent(self, ev) -> bool:
        if not self._pressed:
            return True
        self.update_roi(ev)
        ev.accept()
        return False

    def mouseReleaseEvent(self, ev) -> bool:
        if self._pressed:
            self.update_roi(ev)

            self._init_point: QPointF = None
            self._roi: Roi = None
            self._pressed: bool = False
            ev.accept()
            return False
        return True

    def update_roi(self, ev):
        self._update_roi(self._view_box.mapToView(ev.pos()))
        if self._roi.key is not None:
            self.sigMoveRoi.emit(self._roi.key, '')

    @abstractmethod
    def _update_roi(self, point: QPointF):
        pass

    def _init_roi(self) -> Roi:
        return Roi(0, 0)
